#!/bin/bash
# un-install script for the FireGL2/3 drivers for Linux/x86(IA32) with XFree 4.x

# general variables
MODULE_SPACED="Fire GL 2/3"
MODULE_UPPER=FIREGL23
Module=FireGL23
MODULE=firegl23
MODULE_SHORT=fgl23
family=firegl
family_short=fgl

# general locations
HERE=`pwd`

# xfree locations
XF_ROOT=/usr/X11R6
XF_BIN=$XF_ROOT/bin
XF_LIB=$XF_ROOT/lib
XF_INC=$XF_ROOT/include
XF_MOD=$XF_LIB/modules
XF_DRV=$XF_MOD/drivers
XF_DRI=$XF_MOD/dri
XF_LIN=$XF_MOD/linux
XF_EXT=$XF_MOD/extensions

# system locations
OS_MOD=/lib/modules

# vendor options - tightly related to supplied package contents - do not change!
INSTALL_LIBGL_V2=1

# shell script subroutines
#none

#begin of main script
echo -e "====================================================================="   
echo -e "\033[31m FIRE GL - BE THE CREATOR \033[0m"                               
echo -e "====================================================================="   

#must be executed as root
if ! whoami | grep root > /dev/null ; then
  echo ""
  echo "You must be logged in as 'root' to run this uninstall script."
  echo "Stopping now."
  echo ""
  exit 1
fi

#begin de-installation
doloop=1
while [ $doloop -ne 0 ]; do
  echo ""
  echo -n "Do you want to remove the drivers for the ${MODULE_SPACED} now? (y/n) [y] "
  read response
  if [ ! "$response" != "" ]; then
    response="y" 
  fi
  case "$response" in
    [nN]*)
      exit 0
      ;;
    [yY]*)
      doloop=0
      ;;
    *)
      ;;
  esac
done

#Do pre-installation
#remove old display driver
if [ -f $XF_MOD/${MODULE}_dri.so ]
then
  rm -f $XF_MOD/${MODULE}_dri.so > /dev/null
fi

echo ""
echo "Removing files from working folders..."

# helpers for removing the files from destination location
function remove_src_dst()
{
  rm -f $DST
}

function remove_icd_src_dst()
{
  rm -f $DST
}

# === script and list files, installation helpers ===
#none

# === configuration file generator application ===
SRC=${MODULE}config
DST=$XF_BIN/$SRC
remove_src_dst

# === test / check programs ===
#possibly shared - keep that helper
#SRC=${family}info
#DST=$XF_BIN/$SRC
#remove_src_dst

#possibly shared - keep that helper
#SRC=cpu_check
#DST=$XF_BIN/$SRC
#remove_src_dst

# ========================================

#kernel interface lib
SRC=lib${MODULE_SHORT}.a
DST=$XF_LIN/$SRC
remove_src_dst

#display driver (XServer module)
SRC=${MODULE}_drv.o
DST=$XF_DRV/$SRC
remove_src_dst


#lib and header file for gamma correction extension
SRC=libfgl_gamma.a
DST=$XF_LIB/$SRC
remove_src_dst
SRC=libfgl_gamma.a.1
DST=$XF_LIB/$SRC
remove_src_dst
SRC=libfgl_gamma.a.1.0
DST=$XF_LIB/$SRC
remove_src_dst

SRC=fgl_gamma.h
DST=$XF_INC/X11/extensions/$SRC
remove_src_dst

#gamma correction demo application
SRC=fgl_xgamma
DST=$XF_BIN/$SRC
remove_src_dst

#pbuffer demo application
SRC=fgl_glxgears
DST=$XF_BIN/$SRC
remove_src_dst

#archive with sources for gamma and pbuffer demos
SRC=$(MODULE)sample_source.tgz
DST=/usr/src/$SRC
remove_src_dst


# Pentium optimized ICD
SRC=${MODULE}_dri.so
DST=$XF_DRI/$SRC
remove_src_dst

# remember this:
PENTIUM_ICD=$DST

# AMD Athlon 3DNow! enhanced ICD
SRC=${MODULE}a_dri.so
DST=$XF_DRI/$SRC
remove_icd_src_dst

# Intel Pentium3 ISSE enhanced ICD
SRC=${MODULE}s_dri.so
DST=$XF_DRI/$SRC
remove_icd_src_dst

# Intel Pentium3 optimized ICD _without_ ISSE (kernel support not required)
SRC=${MODULE}ns_dri.so
DST=$XF_DRI/$SRC
remove_icd_src_dst

# Intel Pemtium4 ISSE enhanced ICD
SRC=${MODULE}w_dri.so
DST=$XF_DRI/$SRC
remove_icd_src_dst


# === kernel modules ===
# (some files might refer to files of former driver versions)
# clean up any existing modules and build logs in module subdirs
rm -f `find $OS_MOD/${MODULE}/*/${MODULE}.o 2>/dev/null` 
rm -f `find $OS_MOD/${MODULE}/*/make.log 2>/dev/null`
# clean up module subdirs themselves
rm -f `find $OS_MOD/${MODULE}/2* -type d 2>/dev/null`
# clean up any existing modules and build logs in module dir
rm -f `find $OS_MOD/${MODULE}/${MODULE}*.o 2>/dev/null` 
rm -f `find $OS_MOD/${MODULE}/make*.log 2>/dev/null`
rm -f $OS_MOD/${MODULE}/make_symlink.sh 2>/dev/null
rm -f $OS_MOD/${MODULE}/make_install.sh 2>/dev/null
# clean up any existing kernel build lib files
rm -f `find $OS_MOD/${MODULE}/build_mod/libfgl_ip.*.a 2>/dev/null`
rm -f `find $OS_MOD/${MODULE}/build_mod/libfgl_ip.a 2>/dev/null`
rm -f `find $OS_MOD/${MODULE}/build_mod/lib${MODULE_SHORT}_ip.a 2>/dev/null`
# clean up any existing source or object file
rm -f `find $OS_MOD/${MODULE}/build_mod/*.c 2>/dev/null`
rm -f `find $OS_MOD/${MODULE}/build_mod/*.h 2>/dev/null`
rm -f `find $OS_MOD/${MODULE}/build_mod/*.o 2>/dev/null`

# remove modules directories 

function remove_dir()
{
  if [ -d $DIR ]; then
    rmdir $DIR
  fi
}

function remove_symlink_to_dir()
{
  if [ -h $SYMLINK ]; then
    rm -f $SYMLINK
  fi
}

function remove_symlink_to_file()
{
  if [ -h $SYMLINK ]; then
    rm -f $SYMLINK
  fi
}

# ===============================================================================
# kernel module generator files

# associated info file
SRC=${MODULE}_info.txt       DST=$OS_MOD/${MODULE}/$SRC
remove_src_dst

# kernel module generator file
SRC=make.sh                  DST=$OS_MOD/${MODULE}/build_mod/make.sh
remove_src_dst

# public kernel module source files
SRC=firegl_public.c          DST=$OS_MOD/${MODULE}/build_mod/$SRC
remove_src_dst
SRC=firegl_public.h          DST=$OS_MOD/${MODULE}/build_mod/$SRC
remove_src_dst

SRC=agpgart_fe.c             DST=$OS_MOD/${MODULE}/build_mod/$SRC
remove_src_dst
SRC=agpgart_be.c             DST=$OS_MOD/${MODULE}/build_mod/$SRC
remove_src_dst
SRC=agp_backend.h            DST=$OS_MOD/${MODULE}/build_mod/$SRC
remove_src_dst
SRC=agpgart.h                DST=$OS_MOD/${MODULE}/build_mod/$SRC
remove_src_dst
SRC=agp.h                    DST=$OS_MOD/${MODULE}/build_mod/$SRC
remove_src_dst

SRC=compat-pre24.h           DST=$OS_MOD/${MODULE}/build_mod/$SRC
remove_src_dst
SRC=drm.h                    DST=$OS_MOD/${MODULE}/build_mod/$SRC
remove_src_dst
SRC=drmP.h                   DST=$OS_MOD/${MODULE}/build_mod/$SRC
remove_src_dst
SRC=drm_proc.h               DST=$OS_MOD/${MODULE}/build_mod/$SRC
remove_src_dst


rm -f $OS_MOD/${MODULE}/build_mod/*.o

# remove subdirs
DIR=$OS_MOD/${MODULE}/build_mod
remove_dir

DIR=$OS_MOD/${MODULE}
remove_dir

# ===============================================================================

if [ $INSTALL_LIBGL_V2 -eq 0 ]
then
  # we will keep(!) the new version of libGL.so.1.2
  # nothing to do.
  dummy=0
else
  # we will undo all backuppde executeables of libGL.so*
  # symlink restoring is left to ldconfig and to the admin

  # determine which lib dirs are of relevance in current system
  ldconfig -v -N -X 2>/dev/null | sed -n -e '/^\/.*:$/s/:$//p' >libdirs.txt

  # remove all invalid paths to simplify the following code  
  # have a look for the XF86 lib dir at the same time
  echo -n >libdirs2.txt
  for libdir in `cat libdirs.txt`;
  do
    if [ -d $libdir ]
    then
      echo $libdir >>libdirs2.txt
    fi
  done
  
  # Step 1: broswe all dirs and cleanup existing libGL.so* files
  for libdir in `cat libdirs2.txt`;
  do 
    for libfile in `ls -1 $libdir/libGL.so* 2>/dev/null`;
    do
      libname=`find $libfile -printf %f`
      # act on file, depending on its type
      if [ -h $libdir/$libname ]
      then
        # delete symlinks
        rm -f $libdir/$libname
      else
        if [ -f $libdir/$libname ]
        then
          # remove regular files
          rm -f $libdir/$libname
        else
          echo "WARNING: lib file $libdir/$libname"
          echo "is of unknown type and therefore not handled."
        fi  
      fi
    done 
  done
  
  # Step 2: restore any "libGL.so*" from XFree86
  # - zero sized files will finally get deleted
  for libdir in `cat libdirs2.txt`;
  do 
    for libfile in `ls -1 $libdir/FGL.renamed.libGL.so* 2>/dev/null`;
    do
      libname=`find $libfile -printf %f`
      origlibfile=`echo $libdir/$libname | sed -n -e 's/FGL\.renamed\.//p'`
      origlibname=`echo $libname | sed -n -e 's/FGL\.renamed\.//p'`
      mv $libdir/$libname $libdir/$origlibname
      if [ ! -s $libdir/$origlibname ]
      then
        rm -f $libdir/$origlibname
      fi
    done
  done
  
  # cleanup helper files
  rm -f libdirs.txt libdirs2.txt 
fi


#locate init.d directory
destination=""
if [ -d /etc/rc.d/init.d ]
then
  #must be SuSe or RedHat
  destination="/etc/rc.d"
else
  if [ -d /etc/init.d ]
  then
    #must be Corel or Debian
    destination="/etc"
  fi
fi

if [ $destination != "" ]
then
  rm -f $destination/init.d/${MODULE}
fi

# ==============================================================
echo ""
echo "Processing system configuration..."

# let system rescan library dependencies
echo "- updating library dependencies"
/sbin/ldconfig

#do postinstall actions
#set environment variable LD_LIBRARY_PATH in /etc/profile (bash shell)
FILE=/etc/profile
if ! grep LD_LIBRARY_PATH $FILE | grep $XF_LIB:$XF_MOD > /dev/null
then
  echo "- removing lines from $FILE"
  mv -f $FILE ${FILE}.old
  sed ${FILE}.old -e '/^#BEGIN-'${MODULE_UPPER}'/,/^#END-'${MODULE_UPPER}'/d' >$FILE
fi

#set environment variable LD_LIBRARY_PATH in /etc/csh.cshrc (c shell)
FILE=/etc/csh.cshrc
if ! grep LD_LIBRARY_PATH $FILE | grep $XF_LIB:$XF_MOD >/dev/null
then
  echo "- removing lines from $FILE"
  mv -f $FILE ${FILE}.old
  sed ${FILE}.old -e '/^#BEGIN-'${MODULE_UPPER}'/,/^#END-'${MODULE_UPPER}'/d' >$FILE
fi         

#add kernel module to runlevels 1-5
#locate init.d directory - is already done

if [ $destination != "" ]
then
  echo "- removing ${MODULE} startup service from init.d system"
  
  rm -f $destination/rc1.d/K20${MODULE}
  rm -f $destination/rc1.d/S20${MODULE}
  rm -f $destination/rc2.d/K20${MODULE}
  rm -f $destination/rc2.d/S20${MODULE}
  rm -f $destination/rc3.d/K20${MODULE}
  rm -f $destination/rc3.d/S20${MODULE}
  rm -f $destination/rc4.d/K20${MODULE}
  rm -f $destination/rc4.d/S20${MODULE}
  rm -f $destination/rc5.d/K20${MODULE}
  rm -f $destination/rc5.d/S20${MODULE}
fi

# remove any present init.d start/kill settings
rm -f `ls $destination/rc?.d/{S,K}??${MODULE} 2>/dev/null` >/dev/null
# remove previous init.d scripts
rm -f `ls $destination/${MODULE} 2>/dev/null` >/dev/null
    
# stop any running (and unused) kernel module
module_str=`/sbin/lsmod | grep $MODULE | cut -f 1 -d " " `
if [ "$module_str" = "$MODULE" ];
then
  echo "- removing old but currenly active kernel module $MODULE"
  /sbin/rmmod $MODULE >/dev/null
  exit_val=$?
  if [ $exit_val -ne 0 ];
  then
    echo "Warning: unable to stop this kernel module"
  fi
fi

# for RedHat Linux (and possibly other distirbuions)
# remove the symlink to the config file
if [ -h /etc/X11/XF86Config-4 ];
then
  rm -f /etc/X11/XF86Config-4
fi

echo ""
echo "It is recommended to logoff and logon again at any shell"
echo "in order to update the value of the LD_LIBRARY_PATH variable."
echo ""

exit 0
# EOF
