#!/bin/bash
# install script for the FireGL2/3 drivers for Linux/x86(IA32) with XFree 4.x

# general variables
MODULE_SPACED="Fire GL 2/3"
MODULE_UPPER=FIREGL23
Module=FireGL23
MODULE=firegl23
MODULE_SHORT=fgl23
family=firegl
family_short=fgl

# general locations
HERE=`pwd`
XXX="$HERE"/XXX

i=0
i=$(($i+`echo "$HERE" | grep -c \ `))
i=$(($i+`echo "$HERE" | grep -c \*`))
i=$(($i+`echo "$HERE" | grep -c \\\\\\\\`))
i=$(($i+`echo "$HERE" | grep -c \"`))
i=$(($i+`echo "$HERE" | grep -c \'`))
i=$(($i+`echo "$HERE" | grep -c \\\``))
if [ $i -ne 0 ];
then
  echo "Path to current working directory contains critical caracters"
  echo "like asterisk, backslash, quotes or spaces. Since some system"
  echo "tools wont cope with that situation this script will abort now."
  exit 1
fi


# xfree locations
XF_ROOT=/usr/X11R6
XF_BIN=$XF_ROOT/bin
XF_LIB=$XF_ROOT/lib
XF_INC=$XF_ROOT/include
XF_MOD=$XF_LIB/modules
XF_DRV=$XF_MOD/drivers
XF_DRI=$XF_MOD/dri
XF_LIN=$XF_MOD/linux
XF_EXT=$XF_MOD/extensions

# system locations
OS_MOD=/lib/modules
USR_LIB=/usr/lib

# list of dirs that are used for verifying xfree 4.x presence
XFREEDIRS="$XF_BIN \
           $XF_LIN \
           $XF_DRV \
           $XF_EXT \
	  "

# list of dirs that are used for verifying xfree 4.x presence
XFREEFILES="$XF_BIN/XFree86 \
           "

# vendor options - tightly related to supplied package contents - do not change!
INSTALL_LIBGL=0
INSTALL_LIBGL_V2=1
INSTALL_MODULE_SUBDIRS=0
INSTALL_INIT_SERVICE=0
INSTALL_MODULE_MAKE_INSTALL=1
INSTALL_LIBPATH=0

# shell script defaults
OPTION_FORCE=0
OPTION_QUERY=1


# shell script subroutines
GetOsInfo()
{
	echo "Checking which OS you're running..."

	OS_ame="`uname`"
	OsVersion="`uname -r`"
	case "$OS_ame" in
	SunOS) # Assumes SunOS 5.x
		OsArch="`uname -p`"
		;;
	*)
		OsArch="`uname -m`"
		;;
	esac
	# Some SVR4.0 versions have a buggy uname that reports the node name
	# for the OS name.  Try to catch that here.  Need to check what is
	# reported for non-buggy versions.
	if [ "$OS_ame" = "`uname -n`" -a -f /stand/unix ]; then
		OS_ame=UNIX_SV
	fi
	echo "uname reports '$OS_ame' version '$OsVersion', architecture '$OsArch'."

	# Find the object type, where needed

	case "$OS_ame" in
	Linux|FreeBSD|NetBSD)
		if file -L /bin/sh | grep ELF > /dev/null 2>&1; then
			OsObjFormat=ELF
		else
			OsObjFormat=a.out
		fi
		;;
	esac

	if [ X"$OsObjFormat" != X ]; then
		echo "Object format is '$OsObjFormat'."
		needNL=YES
	fi

	# test's flag for symlinks
	#
	# For OSs that don't support symlinks, choose a type that is guaranteed to
	# return false for regular files and directories.

	case "$OS_ame" in
	FreeBSD)
		case "$OsVersion" in
		2.*)
			L="-h"
			;;
		*)
			L="-L"
		esac
		;;
	OS-with-no-symlinks)	# Need to set this correctly
		L="-b"
		NoSymlinks=YES
		;;
	*)
		L="-L"
		;;
	esac

	# Find the libc version, where needed
	case "$OS_ame" in
	Linux)
		tmp="`ldd /bin/sh | grep libc.so 2> /dev/null`"
		LibcPath=`expr "$tmp" : '[^/]*\(/[^ ]*\)'`
		tmp="`strings $LibcPath | grep -i 'c library'`"
		OsLibcMajor=`expr "$tmp" : '.* \([0-9][0-9]*\)'`
		OsLibcMinor=`expr "$tmp" : '.* [0-9][0-9]*\.\([0-9][0-9]*\)'`
		case "$OsLibcMajor" in
		2)
			# 2 is the glibc version
			OsLibcMajor=6
			;;
		esac
		;;
	esac

	if [ X"$OsLibcMajor" != X ]; then
		if [ X"$OsLibcMinor" != X ]; then
			echo "libc version is '$OsLibcMajor.$OsLibcMinor'."
		else
			echo "libc version is '$OsLibcMajor'."
		fi
		needNL=YES
	fi
	if [ X"$needNL" = XYES ]; then
		echo ""
	fi
#	echo ""
}

GetXfreeInfo()
{
    if [ $chk_env -eq 0 ]
    then
      FILE=$XF_BIN/XFree86
    else
      FILE=../../xc/programs/Xserver/XFree86
    fi

    x_ver_line=`$FILE -version 2>&1 | grep "XFree86 Version"`
    x_ver=`echo "$x_ver_line" | cut -d' ' -f3`

    if [ "$x_ver" = "" ]
    then
      echo "Unable to determine XFree86 Version. Stopping now."
      echo ""
      exit 1
    fi

    if [ $chk_env -eq 0 ]
    then
      echo "XFree86 Version $x_ver is installed."
    else
      echo "XFree86 Version $x_ver was found in your current source environment."
    fi
    echo ""

    # some post processing of our main result
    X_4_0=0
    X_4_1=0
    X_good=0
    if [ `echo "$x_ver" | grep -c '^4\.0\.[23]$'` -gt 0 ]
    then
      X_4_0=1
      X_good=1
    fi
    if [ `echo "$x_ver" | grep -c '^4\.0\.99'` -gt 0 ]
    then
      X_4_1=1
      X_good=1
    fi
    if [ `echo "$x_ver" | grep -c '^4\.1\.0'` -gt 0 ]
    then
      X_4_1=1
      X_good=1
    fi
}

FindPackageName()
{
	case "$OS_ame" in
	DGUX)	# Check this string
		case "$OsArch" in
		i*86)
			Message="DGUX binaries are not available"
			;;
		*)
			Message="DGUX binaries are not available"
			;;
		esac
		;;
	FreeBSD)
		case "$OsArch" in
		i386)
			case "$OsVersion" in
			2.2*)
			        Message="FreeBSD-2.2.x binaries are not available"
				;;
			3.*)
				case "$OsObjFormat" in
				ELF)
			                Message="FreeBSD-3.x binaries are not available"
					;;
				*)
			                Message="FreeBSD-3.x binaries are not available"
					;;
				esac
				;;
			4.*)
			        Message="FreeBSD-4.x binaries are not available"
				;;
			*)
				Message="FreeBSD/i386 binaries are not available"
				;;
			esac
			;;
		alpha)
			case "$OsVersion" in
			3.*)
			        Message="FreeBSD-alpha-3.x binaries are not available"
				;;
			4.*)
			        Message="FreeBSD-alpha-4.x binaries are not available"
				;;
			*)
				Message="FreeBSD/alpha binaries are not available"
				;;
			esac
			;;
		*)
			Message="FreeBSD binaries are not available"
			;;
		esac
		;;
	Linux)
        x_package_suffix=""
        if [ $X_4_0 -gt 0 ] 
        then
            x_package_suffix="-X40"
        fi
        if [ $X_4_1 -gt 0 ] 
        then
            x_package_suffix="-X41"
        fi

		case "$OsArch" in
		i*86)
			case "$OsLibcMajor" in
			5)
				PackageName="flibc5.tgz"
				;;
			6)
				case "$OsLibcMinor" in
				0)
					echo "No package available for glibc 2.$OsLibcMinor.  Try ${MODULE_SHORT}glibc21$x_package_suffix.tgz"
                                        PackageName="${MODULE_SHORT}glibc21$x_package_suffix.tgz"
					;;
				1)
				    PackageName="${MODULE_SHORT}glibc21$x_package_suffix.tgz"
					;;
				2)
				    PackageName="${MODULE_SHORT}glibc22$x_package_suffix.tgz"
					;;
				*)
					echo "No package available for glibc 2.$OsLibcMinor.  Try ${MODULE_SHORT}glibc21$x_package_suffix.tgz"
                                        PackageName="${MODULE_SHORT}glibc21$x_package_suffix.tgz"
					;;
				esac
				;;
			*)
				case "$OsObjFormat" in
				a.out)
					Message="Linux a.out is no longer supported"
					;;
				*)
					Message="No Linux/ix86 binaries for this libc version"
					;;
				esac
				;;
			esac
			;;
		alpha)
			case "$OsLibcMajor.$OsLibcMinor" in
			6.1)
				Message="Linux/alpha binaries are not available"
				;;
			6.*)
				Message="Linux/alpha binaries are not available"
				;;
			*)
				Message="Linux/alpha binaries are not available"
				;;
			esac
			;;
		*)
			Message="No Linux binaries available for this architecture"
			;;
		esac
		;;
	LynxOS)	# Check this
		Message="LynxOS binaries are not available"
		;;
	NetBSD)
		case "$OsArch" in
		i386)
			case "$OsVersion" in
			1.[3-9]*)	# Check this
				case "$OsObjFormat" in
				a.out)
					Message="NetBSD binaries are not available"
					;;
				*)
					Message="NetBSD binaries are not available"
					;;
				esac
				;;
			*)
				Message="NetBSD binaries are not available"
				;;
			esac
			;;
		*)
			Message="NetBSD binaries are not available"
			;;
		esac
		;;
	OpenBSD)
		case "$OsArch" in
		i386)
			case "$OsVersion" in
			2.[6-9]*)	# Check this
				Message="OpenBSD binaries are not available"
				;;
			*)
				Message="OpenBSD binaries are not available"
				;;
			esac
			;;
		*)
			Message="OpenBSD binaries are not available"
			;;
		esac
		;;
	SunOS)
		case "$OsArch" in
		i386)
			case "$OsVersion" in
			5.[67]*)
				Message="SunOS binaries are not available"
				;;
			5.8*)
				Message="SunOS binaries are not available"
				;;
			*)
				Message="SunOS binaries are not available"
				;;
			esac
			;;
		*)
			Message="SunOS binaries are not available"
			;;
		esac
		;;
	UNIX_SV)
		case "$OsArch" in
		i386)
			case "$OsVersion" in
			4.0*)
				Message="SYSV binaries are not available"
				;;
			*)
				# More detailed version check??
				Message="SYSV binaries are not available"
				;;
			esac
			;;
		*)
			Message="SYSV binaries are not available"
			;;
		esac
		;;
	*)
		Message="No binaries available for this OS"
		;;
	esac

	if [ X"$PackageName" != X ]; then
		echo "Binary package name is '$PackageName'"
	else
		if [ X"$Message" = X ]; then
			echo "Can't find which package you should use."
		else
			echo "$Message"
		fi
	fi
    echo ""
}

#begin of main script
echo -e "====================================================================="   
echo -e "\033[31m FIRE GL - BE THE CREATOR \033[0m"                               
echo -e "====================================================================="   

# test for commandline arguments
chk_env=0
if [ "$1" = "-check-environment" ]; then
  chk_env=1
fi

if [ "$1" = "--force" ]; then
  echo "Option 'force' detected."
  OPTION_FORCE=1
  # implies -check-environment
  chk_env=1
fi


# perform a basic system analysis
GetOsInfo
GetXfreeInfo
FindPackageName


# react on command name and commandline arguments
command_name_lower=`find $0 -printf %f | tr [[:upper:]] [[:lower:]]`
if [ "$command_name_lower" = "check.sh" ]; then
  # in check mode only the required libc package name is reported
  exit 0
fi

if [ "$1" = "-check" ]; then
  # in check mode only the required libc package name is reported
  exit 0
fi

if [ "$1" = "--check" ]; then
  # in check mode only the required libc package name is reported
  exit 0
fi

if [ "$1" = "-check-environment" ]; then
  # in check mode only the required libc package name is reported
  exit 0
fi

if [ $OPTION_FORCE -ne 0 ]; then
  if [ ! "$2" = "" ]; then
    echo "Commandline override for package name: $PackageName"
  fi
fi

if [ "$3" = "--noquery" ]; then
  echo "Option 'noquery' detected."
  OPTION_QUERY=0
fi


# veriyfy some settings
if [ "$PackageName" = "" ]; then
  # could not resolve package name
  exit 1
fi

#check, if some xfree86 4.0 or compatible is installed
for i in $XFREEDIRS ; do
  if [ ! -d $i ]; then
    echo ""
    echo "You must have installed XFree86 version 4.0 before."
    echo "Stopping now."
    echo ""
    exit 1
  fi
done

for i in $XFREEFILES ; do
  if [ ! -f $i ]; then
    echo ""
    echo "You must have installed XFree86 version 4.0 before."
    echo "Stopping now."
    echo ""
    exit 1
  fi
done

#check, if we do have the matching drivers and libGL for the 
#currently installed x-server version
if [ $X_good -eq 0 ]
then
  echo ""
  echo "At present drivers are only availabel for releases"
  echo "4.0.2, 4.0.3 and 4.1.0 of XFree86. Stopping now."
  echo "" 
  exit 1
fi

# after general requirements, check real conditions for an install
#must be executed as root
if [ `id -u` -ne 0 ]; then
  echo ""
  echo "You must be logged in as 'root' to run this installation."
  echo "Stopping now."
  echo ""
  exit 1
fi

# try to adjust shell paths if not a login shell
which depmod >/dev/null 2>&1
if [ $? -ne 0 ];
then
#  echo "(completing current path to be a root path)"
#  echo ""
  PATH=/usr/local/sbin:/usr/sbin:/sbin:${PATH}
  which depmod >/dev/null 2>&1
  if [ $? -ne 0 ];
  then
    echo ""
    echo "You arent running in a 'login shell'."
    echo "Please login directly from a console"
    echo "or use 'su -l' to get the same result."
    echo "Stopping now."
    echo ""
    exit 1
  fi
fi


#must have write access to current directory
echo "XXX" > .echotmp 2> /dev/null
if [ ! -f .echotmp ]; then
  echo ""
  echo "Can't write to the current directory. Stopping now."
  echo ""
  exit 1
fi
rm -f .echotmp

#check, if binary package is present
if [ ! -f $PackageName ]; then
  echo ""
  echo "Binary package '$PackageName' is not available. Stopping now."
  echo ""
  exit 1
fi


if [ $OPTION_QUERY -ne 0 ]; then

#ask if we should begin installation now
doloop=1
while [ $doloop -ne 0 ]; do
  echo ""
  echo -n "Do you want to install the drivers for the ${MODULE_SPACED} now? (y/n) [y] "
  read response
  if [ ! "$response" != "" ]; then
    response="y" 
  fi
    case "$response" in
    [nN]*)
      exit 0
      ;;
    [yY]*)
      doloop=0
      ;;
    *)
      ;;
  esac
done

fi

#Do pre-installation
#remove old driver
if [ -f $XF_MOD/${MODULE}_dri.so ]
then
  rm -f $XF_MOD/${MODULE}_dri.so > /dev/null
fi

#expand files from archive

echo ""
echo "Extracting files:"

if [ ! -d $XF_BIN ]; then
  echo ""
  echo "Directory '$XF_BIN' does not exist. Stopping now."
  echo ""
  exit 1
fi

mkdir "$XXX" 2>/dev/null
if [ ! -d "$XXX" ]; then
  echo ""
  echo "Unable to create temporary extraction directory. Stopping now."
  echo ""
  exit 1
fi
cd "$XXX"

# totally cleanup that directory with all possible subdirs
rm -f -R *
# extract whole archive into temp dir in a single step
tar -z -v -x -f ../$PackageName

echo ""
echo "Installing files into working folders:"

# helpers for moving the extracted file into destination location
function extract_src_tmp()
{
#  rm -f $SRC
#  tar -z -v -x $SRC -f ../$PackageName
  if [ ! -f $SRC ]; then
    echo ""
    echo "ERROR: can't extract '$SRC'. Stopping now."
    echo ""
    exit 1
  fi
}

function extract_src_dst()
{
#  rm -f $SRC
#  tar -z -v -x $SRC -f ../$PackageName
  if [ ! -f $SRC ]; then
    echo ""
    echo "ERROR: can't extract '$SRC'. Stopping now."
    echo ""
    exit 1
  fi
  rm -f $DST
  cp $SRC $DST
  chown root $DST
  chgrp root $DST
  chmod 755 $DST
  rm -f $SRC
}

function extract_icd_src_dst()
{
#  rm -f $SRC
#  tar -z -v -x $SRC -f ../$PackageName 2>/dev/null
  if [ ! -f $SRC ]; then
    # this OpenGL ICD was not included in package
    # build a symbolic link to Pentium ICD instead
    rm -f $DST
    ln -s ${PENTIUM_ICD} $DST
  else
    rm -f $DST
    cp $SRC $DST
    chown root $DST
    chgrp root $DST
    chmod 755 $DST
    rm -f $SRC
  fi
}

# === script and list files, installation helpers ===
SRC=file_list.txt
extract_src_tmp

# === configuration file generator application ===
SRC=${MODULE}config
DST=$XF_BIN/$SRC
extract_src_dst

# === test / check programs ===
SRC=${family}info
DST=$XF_BIN/$SRC
extract_src_dst

SRC=cpu_check
DST=$XF_BIN/$SRC
extract_src_dst

# ========================================
#kernel interface lib
if [ ! -d $XF_LIN ]; then
  echo ""
  echo "Directory '$XF_LIN' does not exist. Stopping now."
  echo ""
  exit 1
fi

SRC=lib${MODULE_SHORT}.a
DST=$XF_LIN/$SRC
extract_src_dst


#display driver (XServer module)
if [ ! -d $XF_DRV ]; then
  echo ""
  echo "Directory '$XF_DRV' does not exist. Stopping now."
  echo ""
  exit 1
fi

SRC=${MODULE}_drv.o
DST=$XF_DRV/$SRC
extract_src_dst


#gamma correction support lib and header file (for developer purposes)
SRC=libfgl_gamma.a
DST_DIR=$XF_LIB
DST=$DST_DIR/$SRC
if [ ! -d $DST_DIR ]; then
  mkdir -p $DST_DIR
  chmod 755 $DST_DIR
fi
extract_src_dst

SRC=fgl_gamma.h
DST_DIR=$XF_INC/X11/extensions
DST=$DST_DIR/$SRC
if [ ! -d $DST_DIR ]; then
  mkdir -p $DST_DIR
  chmod 755 $DST_DIR
fi
extract_src_dst

#gamma correction demo application
SRC=fgl_xgamma
DST_DIR=$XF_BIN
DST=$DST_DIR/$SRC
if [ ! -d $DST_DIR ]; then
  mkdir -p $DST_DIR
  chmod 755 $DST_DIR
fi
extract_src_dst

#pbuffer demo application
SRC=fgl_glxgears
DST_DIR=$XF_BIN
DST=$DST_DIR/$SRC
if [ ! -d $DST_DIR ]; then
  mkdir -p $DST_DIR
  chmod 755 $DST_DIR
fi
extract_src_dst

#source code and makefiles of demo applications (for developer purposes)
SRC=${MODULE}sample_source.tgz
DST_DIR=/usr/src
DST=$DST_DIR/$SRC
if [ ! -d $DST_DIR ]; then
  mkdir -p $DST_DIR
  chmod 755 $DST_DIR
fi
extract_src_dst


#opengl ICD files, implemented as DRI modules
if [ ! -d $XF_DRI ]; then
  mkdir $XF_DRI
  chmod 755 $XF_DRI
fi

# Pentium optimized ICD
SRC=${MODULE}_dri.so
DST=$XF_DRI/$SRC
extract_src_dst

# remember this:
PENTIUM_ICD=$DST

# AMD Athlon 3DNow! enhanced ICD
SRC=${MODULE}a_dri.so
DST=$XF_DRI/$SRC
extract_icd_src_dst

# Intel Pentium3 ISSE enhanced ICD
SRC=${MODULE}s_dri.so
DST=$XF_DRI/$SRC
extract_icd_src_dst

# Intel Pentium3 optimized ICD _without_ ISSE (kernel support not required)
SRC=${MODULE}ns_dri.so
DST=$XF_DRI/$SRC
extract_icd_src_dst

# Intel Pemtium4 ISSE enhanced ICD
SRC=${MODULE}w_dri.so
DST=$XF_DRI/$SRC
extract_icd_src_dst


# === kernel modules ===
# clean up any existing modules and build logs in module subdirs
rm -f `find $OS_MOD/${MODULE}/*/${MODULE}.o 2>/dev/null` 
rm -f `find $OS_MOD/${MODULE}/*/make.log 2>/dev/null`
# clean up module subdirs themselves
rmdir `find $OS_MOD/${MODULE}/2* -type d 2>/dev/null`  2>/dev/null
# clean up any existing modules and build logs in module dir
rm -f `find $OS_MOD/${MODULE}/${MODULE}*.o 2>/dev/null` 
rm -f `find $OS_MOD/${MODULE}/make*.log 2>/dev/null`
rm -f $OS_MOD/${MODULE}/make_symlink.sh 2>/dev/null
rm -f $OS_MOD/${MODULE}/make_install.sh 2>/dev/null
# clean up any existing kernel build lib files
rm -f `find $OS_MOD/${MODULE}/build_mod/libfgl_ip.*.a 2>/dev/null`
rm -f `find $OS_MOD/${MODULE}/build_mod/libfgl_ip.a 2>/dev/null`
rm -f `find $OS_MOD/${MODULE}/build_mod/lib${MODULE_SHORT}_ip.a 2>/dev/null`
# clean up any existing source or object file
rm -f `find $OS_MOD/${MODULE}/build_mod/*.c 2>/dev/null`
rm -f `find $OS_MOD/${MODULE}/build_mod/*.h 2>/dev/null`
rm -f `find $OS_MOD/${MODULE}/build_mod/*.o 2>/dev/null`

# make modules directories 

function make_dir()
{
  if [ ! -d $DIR ]; then
    mkdir $DIR
    chmod 755 $DIR
  fi
}

function make_symlink_to_dir()
{
  if [ ! -e $SYMLINK ]; then
    ln -s $DIR $SYMLINK
  fi
}

function make_symlink_to_file()
{
  if [ ! -e $SYMLINK ]; then
    ln -s $FILE $SYMLINK
  fi
}

DIR=$OS_MOD
make_dir

DIR=$OS_MOD/${MODULE}
make_dir

# precompiled kernel modules
for FILE in `cat file_list.txt | grep "${MODULE}[.][[:digit:]]*.*[.]o"`;
do
  SRC=${FILE}                     DST=$OS_MOD/${MODULE}/${SRC}
  extract_src_dst
done

# associated info file
SRC=${MODULE}_info.txt          DST=$OS_MOD/${MODULE}/$SRC
extract_src_dst

# ===
if [ $INSTALL_MODULE_MAKE_INSTALL -eq 1 ]
then

  # extract kernel module install script
  SRC=make_install.sh             DST=$OS_MOD/${MODULE}/${SRC}
  rm -f $DST
  extract_src_dst

  mv $DST $DST.presets  
  #insert a few variable lines into that script
  cat $DST.presets | sed -e "
/^#END-CUSTOM/i\\
module=${MODULE}

/^#END-CUSTOM/i\\
XF_BIN=${XF_BIN}

/^#END-CUSTOM/i\\
OS_MOD=${OS_MOD}

/^#END-CUSTOM/i\\
INSTALL_MODULE_SUBDIRS=${INSTALL_MODULE_SUBDIRS}
  " > $DST
  #end of sed script
  rm -f $DST.presets
  #since we setup this file by meance of pipelining, 
  #it already has the local status for user/group=root/root
  chmod 755 $DST

fi
# ===

# ===============================================================================
# kernel module generator files

# precompiled kernel libs
DIR=$OS_MOD/${MODULE}/build_mod
make_dir

for FILE in `cat file_list.txt | grep lib${MODULE_SHORT}_ip`;
do
  SRC=$FILE       DST=$OS_MOD/${MODULE}/build_mod/$FILE
  extract_src_dst
done

# kernel module generator file
SRC=make.sh                  DST=$OS_MOD/${MODULE}/build_mod/make.sh
extract_src_dst

# customize this file to makefile settings by inserting a few lines
mv ${DST} ${DST}.presets
cat ${DST}.presets | sed -e "
/^#END-CUSTOM/i\\
INSTALL_MODULE_SUBDIRS=${INSTALL_MODULE_SUBDIRS}
" > ${DST}
#end of sed script
rm -f ${DST}.presets
chown root $DST
chgrp root $DST
chmod 755 $DST

# public kernel module source files

SRC=firegl_public.c          DST=$OS_MOD/${MODULE}/build_mod/$SRC
extract_src_dst
SRC=firegl_public.h          DST=$OS_MOD/${MODULE}/build_mod/$SRC
extract_src_dst

#SRC=agpgart_fe.c             DST=$OS_MOD/${MODULE}/build_mod/$SRC
#extract_src_dst
SRC=agpgart_be.c             DST=$OS_MOD/${MODULE}/build_mod/$SRC
extract_src_dst
SRC=agp_backend.h            DST=$OS_MOD/${MODULE}/build_mod/$SRC
extract_src_dst
SRC=agpgart.h                DST=$OS_MOD/${MODULE}/build_mod/$SRC
extract_src_dst
SRC=agp.h                    DST=$OS_MOD/${MODULE}/build_mod/$SRC
extract_src_dst

SRC=compat-pre24.h           DST=$OS_MOD/${MODULE}/build_mod/$SRC
extract_src_dst

SRC=drm.h                    DST=$OS_MOD/${MODULE}/build_mod/$SRC
extract_src_dst

SRC=drmP.h                   DST=$OS_MOD/${MODULE}/build_mod/$SRC
extract_src_dst

SRC=drm_proc.h               DST=$OS_MOD/${MODULE}/build_mod/$SRC
extract_src_dst

# ===============================================================================

# cover SuSE spcial case...
if [ `ls -1 /usr/X11R6/bin/switch2* 2>/dev/null | grep "" -c 2>/dev/null` -gt 0 ]
then
  if [ -e /usr/X11R6/bin/switch2xf86-4 ]
  then
    /usr/X11R6/bin/switch2xf86-4
  fi

  if [ -e /usr/X11R6/bin/switch2xf86_glx ]
  then
    echo "swiching OpenGL library support to XFree86 4.x.x DRI method"
    /usr/X11R6/bin/switch2xf86_glx
  else
    echo "Warning:"
    echo "can't switch OpenGL library support to XFree86 4.x.x DRI method"
    echo "because package xf86_glx-4.*.i386.rpm is not installed."
    echo "please install and run switch2xf86_glx afterwards."
  fi
fi

# ===
if [ $INSTALL_LIBGL -eq 1 ];
then

# install new version of libGL.so.1.2
#rm -f libGL.so.1.2
#tar -z -v -x libGL.so.1.2 -f ../$PackageName
if [ ! -f libGL.so.1.2 ]; then
  echo ""
  echo "ERROR: can't extract 'libGL.so.1.2'. Stopping now."
  echo ""
  exit 1
fi

# XF_LIB backup interlude start
# rename previous libGL.so.1.2 backup (this is a fix for previous installer script)
if [ -f $XF_LIB/libGL.so.1.2.old ]; then
  # for some reasons ldconfig made the "*.so.1" symlink point to this file, rename it!
  mv $XF_LIB/libGL.so.1.2.old $XF_LIB/libGL.s___o.1.2.old 2>/dev/null
  # get it off this computer, really!
  rm -f $XF_LIB/libGL.so.1.2.old
fi
# backup previous libGL.so.1.2 (or remove that file if there is already a backup)
if [ -f $XF_LIB/libGL.so.1.2 ]; then
  if [ ! -f $XF_LIB/libGL.s___o.1.2.old ]; then
    # on first time install make backup of existing XFree binary
    mv $XF_LIB/libGL.so.1.2 $XF_LIB/libGL.s___o.1.2.old
  else
    # simply delete previous version of libGL.so.1.2
    rm -f $XF_LIB/libGL.so.1.2
  fi
fi
# XF_LIB backup interlude end

rm -f $XF_LIB/libGL.so.1.2
cp libGL.so.1.2 $XF_LIB/.
chown root $XF_LIB/libGL.so.1.2
chgrp root $XF_LIB/libGL.so.1.2
chmod 755 $XF_LIB/libGL.so.1.2
rm -f libGL.so.1.2

# remove any symlinks to libGL.so.1.2 (will remove real files with same name as well)
rm -f $XF_LIB/libGL.so
rm -f $XF_LIB/libGL.so.1
# create a complete new set of symlinks to our new version of libGL
ln -s $XF_LIB/libGL.so.1.2 $XF_LIB/libGL.so
ln -s $XF_LIB/libGL.so.1.2 $XF_LIB/libGL.so.1

fi
# ===

# ===
if [ $INSTALL_LIBGL_V2 -eq 1 ];
then
  # determine which lib dirs are of relevance in current system
  ldconfig -v -N -X 2>/dev/null | sed -n -e '/^\/.*:$/s/:$//p' >libdirs.txt

  # remove all invalid paths to simplify the following code  
  # have a look for the XF86 lib dir at the same time
  found_xf86_libdir=0;
  echo -n >libdirs2.txt
  for libdir in `cat libdirs.txt`;
  do
    if [ -d $libdir ]
    then
      echo $libdir >>libdirs2.txt
      if [ "$libdir" = "$XF_LIB" ];
      then
        found_xf86_libdir=1;
      fi
    fi
  done
  
  # first check if the XF86 lib dir is included before we do much chaos
  if [ $found_xf86_libdir -ne 1 ]
  then
    echo ""
    echo "ERROR: XFree86 libdir is not included in library path. Stopping now."
    echo ""
    exit 1
  fi
  
  # Step 1: browse all dirs and cleanup existing libGL.so* files
  for libdir in `cat libdirs2.txt`;
  do 
    for libfile in `ls -1 $libdir/libGL.so* 2>/dev/null`;
    do
      libname=`find $libfile -printf %f`
      # act on file, depending on its type
      if [ -h $libdir/$libname ]
      then
        # delete symlinks
        rm -f $libdir/$libname
      else
        if [ -f $libdir/$libname ]
        then
          # remove/rename regular files
          # depending on backup file
          if [ -e $libdir/FGL.renamed.$libname ]
          then
            # if already a backup exists, simply delete them
            rm -f $libdir/$libname
          else
            # if there is no backup then perform a backup
            mv $libdir/$libname $libdir/FGL.renamed.$libname
          fi
        else
          echo "WARNING: lib file $libdir/$libname"
          echo "is of unknown type and therefore not handled."
        fi  
      fi
    done 
  done
  # cleanup helper files
  rm -f libdirs.txt libdirs2.txt
  
  # Step 2: install the new "libGL.so*" from XFree86
  # - if no backup file exists, simple create an empty dummy
  
  # install official XFree86 version of libGL.so.1.2
  #rm -f libGL.so.1.2
  #tar -z -v -x libGL.so.1.2 -f ../$PackageName
  if [ ! -f libGL.so.1.2 ];
  then
    echo ""
    echo "ERROR: can't extract 'libGL.so.1.2'. Stopping now."
    echo ""
    exit 1
  fi

  # we dont intend to make backups of our own previously installed files
  # therefore check if there is NO backup and then create a dummy
  if [ ! -f $XF_LIB/FGL.renamed.libGL.so.1.2 ]; 
  then
    touch $XF_LIB/FGL.renamed.libGL.so.1.2
  fi
  
  # now totally ensure that there is NO libGL.so.1.2 anymore
  # this will happen only if $XF_LIB path is not part of ldconfig path
  # or ldconfig was unable to remove the existing file at this location
  if [ -f $XF_LIB/libGL.so.1.2 ];
  then
    rm -f $XF_LIB/libGL.so.1.2
    if [ -f $XF_LIB/libGL.so.1.2 ];
    then
      echo "Non-critical ERROR: unable to remove existing libGL.so.1.2 for"
      echo "replacement with matching version from driver package."
      echo "You might expirience problems with hardware accllerated OpenGL."
    fi
  fi
  
  # finally install our libGL file
  cp libGL.so.1.2 $XF_LIB/.
  chown root $XF_LIB/libGL.so.1.2
  chgrp root $XF_LIB/libGL.so.1.2
  chmod 755 $XF_LIB/libGL.so.1.2
  rm -f libGL.so.1.2

  # cleanup standard symlinks
  rm -f $XF_LIB/libGL.so
  rm -f $XF_LIB/libGL.so.1
  rm -f $USR_LIB/libGL.so
  rm -f $USR_LIB/libGL.so.1

  # create a complete new set of symlinks the new version of libGL.so.1.2
  # ldconfig surely wont build them all
  ln -s $XF_LIB/libGL.so.1.2 $XF_LIB/libGL.so
  ln -s $XF_LIB/libGL.so.1.2 $XF_LIB/libGL.so.1
  ln -s $XF_LIB/libGL.so.1.2 $USR_LIB/libGL.so
  ln -s $XF_LIB/libGL.so.1.2 $USR_LIB/libGL.so.1
                
fi
# ===


#locate init.d directory
destination=""
if [ -d /etc/rc.d/init.d ]
then
  #must be SuSe or RedHat
  destination="/etc/rc.d"
else
  if [ -d /etc/init.d ]
  then
    #must be Corel or Debian
    destination="/etc"
  fi
fi

# ===
if [ $INSTALL_INIT_SERVICE -eq 1 ]
then

if [ $destination != "" ]
then
#  rm -f ${MODULE}
#  tar -z -v -x ${MODULE} -f ../$PackageName
  if [ ! -f ${MODULE} ]; then
    echo ""
    echo "ERROR: can't extract '${MODULE}'. Stopping now."
    echo ""
    exit 1
  fi
  
  rm -f $destination/init.d/${MODULE}
  #generate start-script
  #insert a few variable lines while duplicating
  #(this is done via sections tags in the script file 
  #which are referenced by the below inlined sed script)
  cat ${MODULE} | sed -e "
/^#END-CUSTOM/i\\
module=${MODULE}

/^#END-CUSTOM/i\\
XF_BIN=$XF_BIN

/^#END-CUSTOM/i\\
OS_MOD=$OS_MOD
  " > $destination/init.d/${MODULE}
  #end of sed script
  #since we setup this file by meance of pipelining, 
  #it already has the local status for user/group=root/root
  chmod 755 $destination/init.d/${MODULE}
  rm -f ${MODULE}
fi

fi
# ===
echo "done."

# ==============================================================
echo ""
echo "Processing system configuration:"

# let system rescan library dependencies
echo "- updating library dependencies"
/sbin/ldconfig

#do postinstall actions
if [ $INSTALL_LIBPATH -ne 0 ];
then
  #set environment variable LD_LIBRARY_PATH in /etc/profile (bash shell)
  if ! grep LD_LIBRARY_PATH /etc/profile | grep $XF_LIB:$XF_MOD > /dev/null
  then
    echo "- adding lines to /etc/profile"
    cat >>/etc/profile <<-begin_end

#BEGIN-${MODULE_UPPER}
	# inserted by the ${Module} installer
	if [ \$LD_LIBRARY_PATH ];
	then
	    if ! set | grep LD_LIBRARY_PATH | grep $XF_LIB:$XF_MOD > /dev/null
	    then
	 	LD_LIBRARY_PATH=\$LD_LIBRARY_PATH:$XF_LIB:$XF_MOD
	 	export LD_LIBRARY_PATH
	    fi
	else
	    LD_LIBRARY_PATH=$XF_LIB:$XF_MOD
	    export LD_LIBRARY_PATH
	fi
#END-${MODULE_UPPER}
	
begin_end
  # end of inlined text
  fi

  #set environment variable LD_LIBRARY_PATH in /etc/csh.cshrc (c shell)
  if ! grep LD_LIBRARY_PATH /etc/csh.cshrc | grep $XF_LIB:$XF_MOD >/dev/null
  then
    echo "- adding lines to /etc/csh.cshrc"
    cat >>/etc/csh.cshrc <<-begin_end
#BEGIN-${MODULE_UPPER}
	#
	# Appendix Fire GL
	#
	
	if ( \$?LD_LIBRARY_PATH ) then
	    setenv LD_LIBRARY_PATH $XF_LIB:$XF_MOD:\$LD_LIBRARY_PATH
	else
	    setenv LD_LIBRARY_PATH $XF_LIB:$XF_MOD
	endif
#END-${MODULE_UPPER}
	
begin_end
  # end of inlined text 
  fi
fi         
# if $INSTALL_LIBPATH

# ===
if [ $INSTALL_INIT_SERVICE -eq 1 ]
then

#add kernel module to runlevels 1-5
#locate init.d directory - is already done

if [ $destination != "" ]
then
  echo "- adding ${MODULE} startup service to init.d system"
  
  # if update-rc.d is available, use it to generate symbolic link
  if [ -f /usr/sbin/update-rc.d ]
  then
    /usr/sbin/update-rc.d -f ${MODULE} start 20 1 2 3 4 5 . stop 20 1 2 3 4 5 . > /dev/null
  else
    cd $destination/rc1.d
    ln -f -s ../init.d/${MODULE} K20${MODULE}
    ln -f -s ../init.d/${MODULE} S20${MODULE}
    cd $destination/rc2.d
    ln -f -s ../init.d/${MODULE} K20${MODULE}
    ln -f -s ../init.d/${MODULE} S20${MODULE}
    cd $destination/rc3.d
    ln -f -s ../init.d/${MODULE} K20${MODULE}
    ln -f -s ../init.d/${MODULE} S20${MODULE}
    cd $destination/rc4.d
    ln -f -s ../init.d/${MODULE} K20${MODULE}
    ln -f -s ../init.d/${MODULE} S20${MODULE}
    cd $destination/rc5.d
    ln -f -s ../init.d/${MODULE} K20${MODULE}
    ln -f -s ../init.d/${MODULE} S20${MODULE}
    cd "$XXX"
  fi
  
  # load kernel module immedeately
  echo "- starting new ${MODULE} kernel module"
  $destination/init.d/${MODULE} stop
  $destination/init.d/${MODULE} start
fi

else
  # remove any present init.d start/kill settings
  rm -f `ls $destination/rc?.d/{S,K}??${MODULE} 2>/dev/null` >/dev/null
  # remove previous init.d scripts
  rm -f `ls $destination/${MODULE} 2>/dev/null` >/dev/null
    
  # stop any running (and unused) kernel module
  # (gives a clean environment for X startup)
  module_str=`/sbin/lsmod | grep $MODULE | cut -f 1 -d " " `
  if [ "$module_str" = "$MODULE" ];
  then
    echo "- removing old but currenly active kernel module $MODULE"
    /sbin/rmmod $MODULE >/dev/null
	exit_val=$?
	if [ $exit_val -ne 0 ];
	then
	  echo "Warning: unable to stop this kernel module"
	fi
  fi
fi
# ===

# install kernel modules
CUSTOM_KERNEL_BUILD_STEP=0;
if [ $INSTALL_MODULE_MAKE_INSTALL -eq 1 ]
then
  # install our kernel module at current kernel's module repository
  echo "- installing kernel module into module repository"
  cd $OS_MOD/${MODULE}
  ./make_install.sh >/dev/null
  exit_val=$?
  cd "$XXX"
  if [ $exit_val -ne 0 ];
  then
    echo "*** Warning ***"
    echo "*** There is no matching kernel module in this package. ***"
    echo "*** A manual custom kernel build step needs to be done. ***"
    CUSTOM_KERNEL_BUILD_STEP=1;
  fi
fi

# cleanup extraction intermediate dir
cd "$XXX"
rm -f *

# remove extraction intermediate dir
cd "$HERE"
rmdir "$XXX" 2>/dev/null

# for RedHat Linux (and possibly others) we might have to adjust                        
# the config file name by meance of a symlink
if [ -f /etc/X11/XF86Config-4 ];
then
  if [ ! -f /etc/X11/XF86Config-4.old ];
  then
    mv /etc/X11/XF86Config-4 /etc/X11/XF86Config-4.old
  else
    rm -f /etc/X11/XF86Config-4
  fi
  ln -s /etc/X11/XF86Config /etc/X11/XF86Config-4
fi


if [ $OPTION_QUERY -ne 0 ]; then

# call respective configuration utility
echo ""
echo "To configure your X server, you must run '${MODULE}config'."
doloop=1
while [ $doloop -ne 0 ]; do
  echo ""
  echo -n "Do you want to run '${MODULE}config' now? (y/n) [y] "
  read response
  if [ ! "$response" != "" ]; then
    response="y" 
  fi
  case "$response" in
    [nN]*)
      echo ""
      doloop=0
      ;;
    [yY]*)
      $XF_BIN/${MODULE}config
      doloop=0
      ;;
    *)
      ;;
  esac
done

fi

echo ""
if [ $CUSTOM_KERNEL_BUILD_STEP -ne 0 ];
then
  echo "Installation complete for 2D"
  echo "============================"
  echo "You have successfully installed the 2D driver"
  echo "for the DIAMOND ${MODULE_SPACED}."
  echo ""
  echo "Installation for 3D incomplete (needs manual customer setup action)"
  echo "=============================="
  echo "In order to use hardware accelerated OpenGL"
  echo "you have to build your own customized kernel module"
  echo "by manually preparing a propper kernel include environment"
  echo "and then running the two scripts as mentioned below"
  echo "- 'cd /lib/modules/${MODULE}/build_mod; ./make.sh'"
  echo "- 'cd /lib/modules/${MODULE}; ./make_install.sh'"
  echo "(see readme.txt for details on how to do that.)"
  echo ""
  if [ $OPTION_QUERY -ne 0 ]; then
    echo -n "Press enter key to continue..."
    read
    echo ""
  fi
else
  echo "Installation complete for 2D and for 3D"
  echo "======================================="
  echo "You have successfully installed the full driver"
  echo "for the DIAMOND ${MODULE_SPACED}."
  echo ""
fi

if [ $INSTALL_LIBPATH -ne 0 ];
then
  echo "It is recommended to logoff and logon again at any shell"
  echo "in order to update the value of the LD_LIBRARY_PATH variable."
  echo ""
fi

echo "done."

exit 0
# EOF
