/****************************************************************************
 *																			*
 *  Copyright (c) 1999-2000 FGL Graphics, a division of SONICblue Inc.   	*
 *  Copyright (c) 2001-2002 ATI Technologies, Starnberg, GERMANY            *
 *  All Rights Reserved.													*
 *																			*
 *  This software and its associated documentation contains proprietary,	*
 *  confidential and trade secret information of FGL Graphics,				*
 *  a division of SONICblue Inc.											*
 *																			*
 *  a)  no part may be disclosed, distributed, reproduced, transmitted,		*
 *      transcribed, stored in a retrieval system, adapted or translated	*
 *      in any form or by any means electronic, mechanical, magnetic,		*
 *      optical, chemical, manual or otherwise,								*
 *																			*
 *  and																		*
 *																			*
 *  b)  the recipient is not entitled to discover through reverse			*
 *      engineering or reverse compiling or other such techniques or		*
 *      processes the trade secrets contained therein or in the				*
 *		documentation.														*
 *																			*
 ****************************************************************************/

#ifdef __KERNEL__

#ifndef MODULE
!!! This is not currently supported,
!!! since it requires changes to linux/init/main.c.
#endif /* !MODULE */

#define EXPORT_SYMTAB	1

#include <asm/unistd.h> /* for installing the patch wrapper */
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>

#include <linux/config.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
#include <linux/file.h>
#include <linux/pci.h>
#include <linux/wrapper.h>
#include <linux/wait.h>
#include <linux/miscdevice.h>
#include <linux/smp_lock.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
//#include <linux/signal.h>
#include <asm/io.h>
#include <asm/mman.h>
#include <asm/uaccess.h>
#include <asm/processor.h>
#include <asm/mtrr.h>
#include <asm/delay.h>
#include <asm/irq.h>
#ifdef __AGP__
#include <linux/agp_backend.h>
#endif /* __AGP__ */
#include "agp_backend.h"

#if LINUX_VERSION_CODE < 0x020400
#include "compat-pre24.h"
#endif

#include "firegl_public.h"

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif
#ifndef NULL
#define NULL (void*)0
#endif

#ifdef pte_offset_atomic
#define FGL_ATOMIC_PTE
#endif

#ifdef FGL_LINUX253P1_VMA_API
// Linux 2.5.3-pre1 and compatibles
#define FGL_VMA_API_TYPE        struct vm_area_struct *
#define FGL_VMA_API_NAME        vma
#define FGL_VMA_API_PROTO       FGL_VMA_API_TYPE FGL_VMA_API_NAME,
#define FGL_VMA_API_PASS        FGL_VMA_API_NAME,
#else /* FGL_253P1_VMA_API */
// Linux 2.4.0 and compatibles
#define FGL_VMA_API_TYPE        /* none */
#define FGL_VMA_API_NAME        /* none */
#define FGL_VMA_API_PROTO       /* none */
#define FGL_VMA_API_PASS        /* none */
#endif /* FGL_253P1_VMA_API */

/* globals */
char* firegl = NULL;
int __ke_debuglevel = 0;
int __ke_moduleflags = 0;
int __ke_debflags = 0;

/* global module vars and constants */
MODULE_PARM(firegl, "s");
MODULE_AUTHOR("Fire GL Graphics - ATI Research GmbH, Germany");

#ifdef FIREGL1
MODULE_DESCRIPTION("ATI Fire GL1");
#define FIREGL_KERNEL_DRIVER_NAME "firegl1"     // needed for kernel message macros
#endif
#ifdef FIREGL23
MODULE_DESCRIPTION("ATI Fire GL2/ Fire GL3");
#define FIREGL_KERNEL_DRIVER_NAME "firegl23"    // needed for kernel message macros
#endif

/* globals constants */
const char*         __ke_UTS_RELEASE        = UTS_RELEASE;
const unsigned int  __ke_PAGE_SHIFT         = PAGE_SHIFT;
const unsigned int  __ke_PAGE_SIZE          = PAGE_SIZE;
const unsigned int  __ke_PAGE_MASK          = PAGE_MASK;
const unsigned long __ke_HZ                 = HZ;
const unsigned long __ke_LINUX_VERSION_CODE = LINUX_VERSION_CODE;

// create global constants and hint symbols (i.e. for objdump checking)
#ifdef MODVERSIONS
const unsigned long __ke_MODVERSIONS_State = 1;
const char BUILD_KERNEL_HAS_MODVERSIONS_SET;
#else
const unsigned long __ke_MODVERSIONS_State = 0;
const char BUILD_KERNEL_HAS_MODVERSIONS_CLEARED;
#endif

#ifdef __SMP__
const unsigned long __ke_SMP_State = 1;
const char BUILD_KERNEL_HAS_SMP_SET;
#else
const unsigned long __ke_SMP_State = 0;
const char BUILD_KERNEL_HAS_SMP_CLEARED;
#endif

/* global structures */
static struct file_operations firegl_fops =
{
#ifdef THIS_MODULE
    owner:   THIS_MODULE,
#endif
    open:    firegl_open,
    release: firegl_release,
    ioctl:   firegl_ioctl,
    mmap:    firegl_mmap,
};

typedef struct {
    __ke_device     pubdev;     // MUST BE FIRST MEMBER
    dev_t           device;     // Device number for mknod

    /* Locking */
    spinlock_t      spinlock[__KE_MAX_SPINLOCKS];       /* For inuse, open_count, buf_use   */
    struct semaphore struct_sem[__KE_MAX_SEMAPHORES];   /* For linked list manipulations    */
} device_t;

static device_t     firegl_public_device;   // The Fire GL public device


/*****************************************************************************/
// standard XFree86 DRM proc support

#define DRM(x) FGLDRM_##x

#ifdef FIREGL_410
#include "drm.h"

// mem_info() is missing in drm_proc.h. But, it is a DRM design problem anyway!
// The first registered DRM device could never report memory statisticts of another
// DRM device, cause the DRM mem_info routine uses local variables. So, let's use a dummy.
static int DRM(mem_info)(char *buf, char **start, off_t offset, int len, int *eof, void *data)
{
    *eof = 1;
    return 0;
}
#include "drm_proc.h"
#else
struct drm_proc_list {
	const char *name;
	int	   (*f)(char *, char **, off_t, int, int *, void *);
} DRM(proc_list)[] = {
	{ "",    0    },
};
#define DRM_PROC_ENTRIES 0
#endif
#ifndef DRM_MAJOR
#define DRM_MAJOR       226
#endif // DRM_MAJOR

static __ke_proc_list_t *drm_proclist = NULL;

/*****************************************************************************/
// Fire GL DRM stub support (compatible with standard DRM stub)

#define FIREGL_STUB_MAXCARDS    16

static struct firegl_stub_list_t {
	const char             *name;
	struct file_operations *fops;
	struct proc_dir_entry  *dev_root;
    __ke_proc_list_t       *proclist;
} firegl_stub_list[FIREGL_STUB_MAXCARDS];

static struct proc_dir_entry *firegl_stub_root;
static int firegl_minor;

static struct firegl_stub_info_t {
    int (*info_register)(const char *name, struct file_operations *fops, device_t *dev);
	int (*info_unregister)(int minor);
    unsigned long signature; // to check for compatible Fire GL DRM device
} firegl_stub_info;

#if LINUX_VERSION_CODE < 0x020400
struct firegl_stub_info_t *firegl_stub_pointer = NULL;
#define inter_module_put(x)
#define inter_module_unregister(x)
#define inter_module_get(x)             firegl_stub_pointer
#define inter_module_register(x,y,z)    do { firegl_stub_pointer = z; } while (0)
/* This is a kludge for backward compatibility that is only useful in DRM(stub_open) */
#define fops_put(fops)      MOD_DEC_USE_COUNT
#define fops_get(fops)      (fops); MOD_INC_USE_COUNT
#endif // LINUX_VERSION_CODE < 0x020400

static struct proc_dir_entry *firegl_proc_init( device_t *dev,
                                                int minor,
				                                struct proc_dir_entry *root,
				                                struct proc_dir_entry **dev_root,
                                                __ke_proc_list_t *proc_list ) // proc_list must be terminated!
{
	struct proc_dir_entry *ent;
	char    name[64];
    __ke_proc_list_t *list = proc_list;
    __KE_DEBUG("minor %d, proc_list 0x%08lx\n", minor, (unsigned long)proc_list);

	if (!minor)
        root = create_proc_entry("dri", S_IFDIR, NULL);

	if (!root) {
		__KE_ERROR("Cannot create /proc/dri\n");
		return NULL;
	}

	sprintf(name, "%d", minor);
	*dev_root = create_proc_entry(name, S_IFDIR, root);
	if (!*dev_root) {
		__KE_ERROR("Cannot create /proc/%s\n", name);
		return NULL;
	}

    while (list->f)
    {
		ent = create_proc_entry(list->name, S_IFREG|S_IRUGO, *dev_root);
		if (!ent)
        {
			__KE_ERROR("Cannot create /proc/dri/%s/%s\n", name, list->name);
            while (proc_list != list)
            {
				remove_proc_entry(proc_list->name, *dev_root);
                proc_list++;
            }
			remove_proc_entry(name, root);
			if (!minor)
                remove_proc_entry("dri", NULL);
			return NULL;
		}

		ent->read_proc = (read_proc_t*)list->f;
        ent->data      = (dev->pubdev.signature == FGL_DEVICE_SIGNATURE) ? dev->pubdev.privdev : dev;

        list++;
	}

	return root;
}

static int firegl_proc_cleanup( int minor,
                                struct proc_dir_entry *root,
		                        struct proc_dir_entry *dev_root,
                                __ke_proc_list_t *proc_list )
{
	char name[64];
    __KE_DEBUG("minor %d\n", minor);

	if (!root || !dev_root)
    {
        __KE_ERROR("no root\n");
        return 0;
    }

    while (proc_list->f)
    {
		remove_proc_entry(proc_list->name, dev_root);
        proc_list++;
    }
	sprintf(name, "%d", minor);
	remove_proc_entry(name, root);
	if (!minor)
        remove_proc_entry("dri", NULL);

	return 0;
}

static int firegl_stub_open(struct inode *inode, struct file *filp)
{
	int                    minor = MINOR(inode->i_rdev);
	int                    err   = -ENODEV;
	struct file_operations *old_fops;
    __KE_DEBUG("\n");

	if (!firegl_stub_list[minor].fops)
        return -ENODEV;
	old_fops   = filp->f_op;
	filp->f_op = fops_get(firegl_stub_list[minor].fops);
	if (filp->f_op->open && (err = filp->f_op->open(inode, filp))) {
		fops_put(filp->f_op);
		filp->f_op = fops_get(old_fops);
	}
	fops_put(old_fops);
	return err;
}

static struct file_operations firegl_stub_fops = {
#if LINUX_VERSION_CODE >= 0x020400
	owner:   THIS_MODULE,
#endif
	open:	 firegl_stub_open
};

static int firegl_stub_getminor(const char *name, struct file_operations *fops, device_t *dev)
{
	int i;
    __KE_DEBUG("name=\"%s\"\n", name);
	for (i = 0; i < FIREGL_STUB_MAXCARDS; i++)
    {
		if (!firegl_stub_list[i].fops)
        {
			firegl_stub_list[i].name = name;
			firegl_stub_list[i].fops = fops;
            firegl_stub_list[i].proclist = (dev->pubdev.signature == FGL_DEVICE_SIGNATURE) ? dev->pubdev.proclist : drm_proclist;
            firegl_stub_root = firegl_proc_init(dev, i, firegl_stub_root, &firegl_stub_list[i].dev_root, firegl_stub_list[i].proclist);
            __KE_DEBUG("minor=%d\n", i);
			return i;
		}
	}
    __KE_DEBUG("no more free minor\n");
	return -1;
}

static int firegl_stub_putminor(int minor)
{
    __KE_DEBUG("minor=%d\n", minor);
	if (minor < 0 || minor >= FIREGL_STUB_MAXCARDS)
        return -1;

	firegl_proc_cleanup(minor, firegl_stub_root, firegl_stub_list[minor].dev_root, firegl_stub_list[minor].proclist);
	firegl_stub_list[minor].name = NULL;
	firegl_stub_list[minor].fops = NULL;
    firegl_stub_list[minor].proclist = NULL;

	if (minor) {
		inter_module_put("drm");
	} else {
		inter_module_unregister("drm");
		unregister_chrdev(DRM_MAJOR, "drm");
	}
	return 0;
}

static int firegl_stub_register(const char *name, struct file_operations *fops, device_t *dev)
{
	struct firegl_stub_info_t *i = NULL;
    __KE_DEBUG("name=\"%s\"\n", name);

	if (register_chrdev(DRM_MAJOR, "drm", &firegl_stub_fops))
		i = (struct firegl_stub_info_t *)inter_module_get("drm");

    if (i)
    {
        // drm device was already registered by another vendors DRM module
        // check whether this was a compatible FireGL device
        if (i->signature == FGL_DEVICE_SIGNATURE)
        {
		    firegl_stub_info.info_register   = i->info_register;
		    firegl_stub_info.info_unregister = i->info_unregister;
        }
        else
        {
            __KE_ERROR("Fire GL kernel module has to be loaded prior to any other DRM kernel module!\n");
		    inter_module_put("drm");
            return -1;
        }
    }
	else
    {
		firegl_stub_info.info_register   = firegl_stub_getminor;
		firegl_stub_info.info_unregister = firegl_stub_putminor;
		inter_module_register("drm", THIS_MODULE, &firegl_stub_info);
	}

	if (firegl_stub_info.info_register)
		return firegl_stub_info.info_register(name, fops, dev);

	return -1;
}

static int firegl_stub_unregister(int minor)
{
	__KE_DEBUG("%d\n", minor);
	if (firegl_stub_info.info_unregister)
		return firegl_stub_info.info_unregister(minor);
	return -1;
}


/*****************************************************************************/
/* init_module is called when insmod is used to load the module */
#if LINUX_VERSION_CODE >= 0x020400
static int __init firegl_init_module(void)
#else
int init_module(void)
#endif
{
    device_t* dev = &firegl_public_device;
    int i, retcode;

	EXPORT_NO_SYMBOLS;

    // init DRM proc list
    drm_proclist = kmalloc((DRM_PROC_ENTRIES + 1) * sizeof(__ke_proc_list_t), GFP_KERNEL);
    if ( drm_proclist == NULL )
        return -ENOMEM;

    for ( i=0; i<DRM_PROC_ENTRIES; i++ )
    {
        drm_proclist[i].name = DRM(proc_list)[i].name;
        drm_proclist[i].f = (__ke_read_proc_t)DRM(proc_list)[i].f;
    }
    drm_proclist[i].f = NULL; // terminate list

    memset(&firegl_stub_list, 0, sizeof(firegl_stub_list));
    memset(&firegl_stub_info, 0, sizeof(firegl_stub_info));
    firegl_stub_info.signature = FGL_DEVICE_SIGNATURE;

    memset(dev, 0, sizeof(*dev));
    dev->pubdev.signature = FGL_DEVICE_SIGNATURE;

    for (i = 0; i < __KE_MAX_SPINLOCKS; i++)
        dev->spinlock[i] = SPIN_LOCK_UNLOCKED;

    for (i=0; i < __KE_MAX_SEMAPHORES; i++)
        sema_init(&dev->struct_sem[i], 1);

	if ( (retcode = firegl_init(&dev->pubdev)) )
    {
        kfree(drm_proclist);
        return retcode;
    }

	// get the minor number
	firegl_minor = firegl_stub_register(dev->pubdev.name, &firegl_fops, dev);
	if (firegl_minor < 0)
    {
        kfree(drm_proclist);
		return -EPERM;
    }

	dev->device = MKDEV(DRM_MAJOR, firegl_minor);

    __KE_INFO("module loaded - %s %d.%d.%d [%s] on minor %d\n",
            dev->pubdev.name,
	        dev->pubdev.major_version,
	        dev->pubdev.minor_version,
	        dev->pubdev.patchlevel,
	        dev->pubdev.date,
		    firegl_minor);

    return 0; // OK!
}

/* cleanup_module is called when rmmod is used to unload the module */
#if LINUX_VERSION_CODE >= 0x020400
static void __exit firegl_cleanup_module(void)
#else
void cleanup_module(void)
#endif
{
    device_t* dev = &firegl_public_device;
    __KE_DEBUG("module cleanup started\n");

	if ( firegl_stub_unregister(firegl_minor) ) {
		__KE_ERROR("Cannot unload module\n");
	}

    firegl_cleanup(&dev->pubdev);

    if (drm_proclist)
        kfree(drm_proclist);

	__KE_INFO("module unloaded - %s %d.%d.%d [%s] on minor %d\n",
            dev->pubdev.name,
	        dev->pubdev.major_version,
	        dev->pubdev.minor_version,
	        dev->pubdev.patchlevel,
	        dev->pubdev.date,
		    firegl_minor);

}

#ifdef MODULE_LICENSE
MODULE_LICENSE("Proprietary. (C) 2002 - ATI Technologies, Starnberg, GERMANY");
#endif

#if LINUX_VERSION_CODE >= 0x020400
module_init( firegl_init_module );
module_exit( firegl_cleanup_module );
#endif

/*****************************************************************************/

/* Wait queue declarations changes in 2.3.1 */
#ifndef DECLARE_WAITQUEUE
#define DECLARE_WAITQUEUE(w, c) struct wait_queue w = { c, NULL }
typedef struct wait_queue* wait_queue_head_t;
#define init_waitqueue_head(q) *q = NULL;
#endif

void __ke_init_waitqueue_head(__ke_wait_queue_t* queue)
{
    init_waitqueue_head((wait_queue_head_t*)queue);
}

void __ke_wake_up_interruptible(__ke_wait_queue_t* queue)
{
    wake_up_interruptible((wait_queue_head_t*)queue);
}

void __ke_add_wait_queue(__ke_wait_queue_t* queue, __ke_wait_queue_t entry)
{
#if LINUX_VERSION_CODE >= 0x020400
	((wait_queue_t*)entry)->flags = 0x0;
	((wait_queue_t*)entry)->task = current;
	((wait_queue_t*)entry)->task_list.next = NULL;
	((wait_queue_t*)entry)->task_list.prev = NULL;
#if WAITQUEUE_DEBUG
	((wait_queue_t*)entry)->__magic = &(((wait_queue_t*)entry)->__magic);
	((wait_queue_t*)entry)->__waker = 0;
#endif

    add_wait_queue((wait_queue_head_t*)queue, (wait_queue_t*)entry);
#else /* LINUX_VERSION_CODE < 0x020400 */
    ((struct wait_queue*)entry)->task = current;
    ((struct wait_queue*)entry)->next = NULL;

    add_wait_queue((wait_queue_head_t*)queue, (struct wait_queue*)entry);
#endif /* LINUX_VERSION_CODE < 0x020400 */
}

void __ke_remove_wait_queue(__ke_wait_queue_t* queue, __ke_wait_queue_t entry)
{
//    current->state = TASK_RUNNING;
#if LINUX_VERSION_CODE >= 0x020400
    remove_wait_queue((wait_queue_head_t*)queue, (wait_queue_t*)entry);
#else /* LINUX_VERSION_CODE < 0x020400 */
    remove_wait_queue((wait_queue_head_t*)queue, (struct wait_queue*)entry);
#endif /* LINUX_VERSION_CODE < 0x020400 */
}

void __ke_schedule(void)
{
	schedule();
}

int __ke_signal_pending(void)
{
    return signal_pending(current);
}

void __ke_set_current_state_task_interruptible(void)
{
    current->state = TASK_INTERRUPTIBLE;
}

void __ke_set_current_state_task_running(void)
{
    current->state = TASK_RUNNING;
}

void __ke_configure_sigmask(__ke_sigset_t *pSigMask)
{
    sigemptyset((sigset_t*)pSigMask);
    sigaddset((sigset_t*)pSigMask, SIGSTOP);
    sigaddset((sigset_t*)pSigMask, SIGTSTP);
    sigaddset((sigset_t*)pSigMask, SIGTTIN);
    sigaddset((sigset_t*)pSigMask, SIGTTOU);
}

void __ke_block_all_signals(int (*notifier)(void *priv), void *pPriv,
         __ke_sigset_t *pSigMask)
{
#if LINUX_VERSION_CODE >= 0x020400
    block_all_signals(notifier,pPriv,(sigset_t*)pSigMask);
#endif
}

void __ke_unblock_all_signals(void)
{
#if LINUX_VERSION_CODE >= 0x020400
    unblock_all_signals();
#endif
}

#if defined(__i386__)

#ifndef __HAVE_ARCH_CMPXCHG
static inline 
unsigned long __cmpxchg(volatile void *ptr, unsigned long old,            
                        unsigned long new, int size)                      
{                                                                                       
    unsigned long prev;                                                             
    switch (size) {                                                                 
    case 1:                                                                         
        __asm__ __volatile__(LOCK_PREFIX "cmpxchgb %b1,%2"
                             : "=a"(prev)
                             : "q"(new), "m"(*__xg(ptr)), "0"(old)
                             : "memory");
        return prev;
    case 2:
        __asm__ __volatile__(LOCK_PREFIX "cmpxchgw %w1,%2"
                             : "=a"(prev)
                             : "q"(new), "m"(*__xg(ptr)), "0"(old)
                             : "memory");
        return prev;
    case 4:
        __asm__ __volatile__(LOCK_PREFIX "cmpxchgl %1,%2"
                             : "=a"(prev)
                             : "q"(new), "m"(*__xg(ptr)), "0"(old)
                             : "memory");
        return prev;
    }
    return old;
}
#endif

#elif defined(__alpha__)
todo !!!
#endif

unsigned long __ke__cmpxchg(volatile void *ptr, unsigned long old,
         unsigned long new, int size)
{
    return __cmpxchg(ptr,old,new,size);
}
                      
/*****************************************************************************/

__ke_dev_t __ke_getdevice(__ke_device_t dev)
{
    return ((device_t*)dev)->device;
}

const char* __ke_module_parm(void)
{
    return firegl;
}

/*****************************************************************************/

int __ke_inode_rdev_minor(struct inode* inode)
{
    return MINOR(inode->i_rdev);
}

/*****************************************************************************/

void* __ke_get_file_priv(struct file* filp)
{
    return filp->private_data;
}

void __ke_set_file_priv(struct file* filp, void* private_data)
{
    filp->private_data = private_data;
}

int __ke_file_excl_open(struct file* filp)
{
    return (filp->f_flags & O_EXCL) != 0;
}

int __ke_file_rw_open(struct file* filp)
{
    return (filp->f_flags & 3) != 0;
}

unsigned int __ke_file_counter(struct file *filp)
{
#if LINUX_VERSION_CODE >= 0x020400
    return filp->f_count.counter;
#else		  
    return filp->f_count;
#endif
}

/*****************************************************************************/

int __ke_getpid(void)
{
    return current->pid;
}

int __ke_geteuid(void)
{
    return current->euid;
}

/*****************************************************************************/

unsigned long __ke_jiffies(void)
{
    return jiffies;
}

void __ke_udelay(unsigned long usecs) // delay in usec
{
    unsigned long start;
    unsigned long stop;
    unsigned long period;
    unsigned long wait_period;
    struct timespec tval;

    if (usecs > 10000)
    {
        start = jiffies;
        tval.tv_sec = usecs / 1000000;
        tval.tv_nsec = (usecs - tval.tv_sec * 1000000) * 1000;
        wait_period = timespec_to_jiffies(&tval);
        do {
            stop = jiffies;

            if (stop < start) // jiffies overflow
                period = ((unsigned long)-1 - start) + stop + 1;
            else
                period = stop - start;

        } while (period < wait_period);
    }
    else
        __udelay(usecs);
}

void __ke_mdelay(unsigned long msecs) // delay in msec
{
	mdelay(msecs);
}

/*****************************************************************************/

unsigned long __ke_virt_to_bus(void* address)
{
    return virt_to_bus(address);
}

unsigned long __ke_virt_to_phys(void* address)
{
    return virt_to_phys(address);
}

void* __ke_high_memory(void)
{
    return high_memory;
}

/*****************************************************************************/

int __ke_error_code(enum __ke_error_num errcode)
{
    switch (errcode)
    {
        case __KE_EBUSY:
            return EBUSY;
        case __KE_EINVAL:
            return EINVAL;
        case __KE_EACCES:
            return EACCES;
        case __KE_EFAULT:
            return EFAULT;
        case __KE_EIO:
            return EIO;
        case __KE_EBADSLT:
            return EBADSLT;
        case __KE_ENOMEM:
            return ENOMEM;
        case __KE_EPERM:
            return EPERM;
        case __KE_ENODEV:
            return ENODEV;
        case __KE_EINTR:
            return EINTR;
        case __KE_ERESTARTSYS:
            return ERESTARTSYS;
        case __KE_ELIBBAD:
            return ELIBBAD;
        default:
            return EFAULT;
    }
}

/*****************************************************************************/

void __ke_mod_inc_use_count(void)
{
    MOD_INC_USE_COUNT;
	__KE_VMDEBUG("++ module count\n");
}

void __ke_mod_dec_use_count(void)
{
    MOD_DEC_USE_COUNT;
	__KE_VMDEBUG("-- module count\n");
}

/*****************************************************************************/

void __ke_down_struct_sem(__ke_device_t dev, int index)
{
    down(&(((device_t*)dev)->struct_sem[index]));
}

void __ke_up_struct_sem(__ke_device_t dev, int index)
{
    up(&(((device_t*)dev)->struct_sem[index]));
}

void __ke_sema_init(struct semaphore* sem, int value)
{
    sema_init(sem, value);
}

__ke_size_t __ke_sema_size(void)
{
    return sizeof(struct semaphore);
}

void __ke_down(struct semaphore* sem)
{
    down(sem);
}

void __ke_up(struct semaphore* sem)
{
    up(sem);
}

/*****************************************************************************/

void __ke_atomic_inc(void* v)
{
    atomic_inc((atomic_t*)v);
}

void __ke_atomic_dec(void* v)
{
    atomic_dec((atomic_t*)v);
}

void __ke_atomic_add(int val, void* v)
{
    atomic_add(val, (atomic_t*)v);
}

void __ke_atomic_sub(int val, void* v)
{
    atomic_sub(val, (atomic_t*)v);
}

int __ke_atomic_read(void* v)
{
    return atomic_read((atomic_t*)v);
}

void __ke_atomic_set(void* v, int val)
{
    atomic_set((atomic_t*)v, val);
}

/*****************************************************************************/

void __ke_spin_lock(__ke_device_t dev, int ndx)
{
   	spin_lock(&(((device_t*)dev)->spinlock[ndx]));
}

void __ke_spin_unlock(__ke_device_t dev, int ndx)
{
   	spin_unlock(&(((device_t*)dev)->spinlock[ndx]));
}

void __ke_lock_kernel(void)
{
    lock_kernel();
}

void __ke_unlock_kernel(void)
{
    unlock_kernel();
}

/*****************************************************************************/

extern unsigned long sys_call_table[];

typedef int (*PFNMLOCK)(unsigned long start, __ke_size_t len);
typedef int (*PFNMUNLOCK)(unsigned long start, __ke_size_t len);


int __ke_sys_mlock(unsigned long start, __ke_size_t len)
{
    PFNMLOCK sys_mlock = (PFNMLOCK)sys_call_table[__NR_mlock];
    if (!sys_mlock) {
        __KE_ERROR("sys_call_table[__NR_mlock] == 0\n");
        return -1;
    }

    return (*sys_mlock)(start, len);
}

int __ke_sys_munlock(unsigned long start, __ke_size_t len)
{
    PFNMUNLOCK sys_munlock = (PFNMUNLOCK)sys_call_table[__NR_munlock];
    if (!sys_munlock) {
        __KE_ERROR("sys_call_table[__NR_munlock] == 0\n");
        return -1;
    }

    return (*sys_munlock)(start, len);
}

/*****************************************************************************/

#ifdef __KE_NO_VSPRINTF

#if LINUX_VERSION_CODE >= 0x020400
void __ke_printk(const char* fmt, ...)
{
	char buffer[256];
    va_list marker;

    va_start(marker, fmt);
    vsprintf(buffer, fmt, marker);
    va_end(marker);

	printk(buffer);
}
#else /* LINUX_VERSION_CODE < 0x020400 */
#define sym2str(_sym) #_sym

#define JmpFunction(name, _kernel_name)         \
    __asm__ (                                   \
        ".globl " name "                    \n" \
        ".type  " name ",@function          \n" \
        name ":                             \n" \
        "   jmp " sym2str(_kernel_name) "   \n" \
    )

JmpFunction("__ke_printk", printk);

#endif /* LINUX_VERSION_CODE < 0x020400 */

#else

void __ke_print_info(const char* fmt, ...)
{
    char msg[256] = KERN_INFO;:
    va_list marker;

    va_start(marker, fmt);
    vsprintf(msg + strlen(msg), fmt, marker);
    va_end(marker);
}

void __ke_print_error(const char* fmt, ...)
{
    char msg[256] = KERN_ERR;
    va_list marker;

    va_start(marker, fmt);
    vsprintf(msg + strlen(msg), fmt, marker);
    va_end(marker);
}

void __ke_print_debug(const char* fmt, ...)
{
    char msg[256] = KERN_DEBUG;
    va_list marker;

    va_start(marker, fmt);
    vsprintf(msg + strlen(msg), fmt, marker);
    va_end(marker);
}

#endif

/*****************************************************************************/

int __ke_capable(enum __ke_cap cap)
{
    switch (cap)
    {
        case __KE_CAP_SYS_ADMIN:
            cap = CAP_SYS_ADMIN;
			break;
        case __KE_CAP_IPC_LOCK:
            cap = CAP_IPC_LOCK;
			break;
        default:
            return 0;
    }
    return capable(cap);
}

void __ke_cap_effective_raise(enum __ke_cap cap)
{
    switch (cap)
    {
        case __KE_CAP_SYS_ADMIN:
            cap = CAP_SYS_ADMIN;
			break;
        case __KE_CAP_IPC_LOCK:
            cap = CAP_IPC_LOCK;
			break;
        default:
            return;
    }
    cap_raise(current->cap_effective, cap);
}

__ke_u32 __ke_get_cap_effective()
{
    return cap_t(current->cap_effective);
}

void __ke_set_cap_effective(__ke_u32 cap)
{
    cap_t(current->cap_effective) = cap;
}

unsigned long __ke_ram_available(void)
{
	struct sysinfo si;

    si_meminfo(&si);
#if LINUX_VERSION_CODE < 0x020317
    /* Changed to page count in 2.3.23 */
	return si.totalram >> PAGE_SHIFT;
#else
	return si.totalram;
#endif
}

int __ke_copy_from_user(void* to, const void* from, __ke_size_t size)
{
    return copy_from_user(to, from, size);
}

int __ke_copy_to_user(void* to, const void* from, __ke_size_t size)
{
    return copy_to_user(to, from, size);
}

int __ke_verify_area(int type, const void * addr, unsigned long size)
{
    return verify_area(type, addr, size);
}

int __ke_get_pci_device_info(struct pci_dev *pdevice, __ke_pci_device_info_t *pinfo)
{
    if ( pdevice )
    {
#if LINUX_VERSION_CODE >= 0x020400
        pinfo->subsystem_vendor = pdevice->subsystem_vendor;
        pinfo->subsystem_device = pdevice->subsystem_device;
#else // LINUX_VERSION_CODE >= 0x020400
		u16 subsystem_vendor;
		u16 subsystem_device;

		pci_read_config_word(pdevice, PCI_SUBSYSTEM_VENDOR_ID, &subsystem_vendor); 
		pci_read_config_word(pdevice, PCI_SUBSYSTEM_ID, &subsystem_device); 
		pinfo->subsystem_vendor = subsystem_vendor;
		pinfo->subsystem_device = subsystem_device;
#endif // LINUX_VERSION_CODE >= 0x020400
        pinfo->vendor = pdevice->vendor;
        pinfo->device = pdevice->device;
    }
    return -EINVAL;
}

int __ke_check_pci(int busnum, int devnum, int funcnum, __ke_u16* vendor, __ke_u16* device, unsigned int* irq)
{
    struct pci_dev* pci_dev;

    pci_dev = pci_find_slot(busnum, PCI_DEVFN(devnum, funcnum));
    if (!pci_dev)
        return 0;

    if (vendor)
        *vendor = pci_dev->vendor;

    if (device)
        *device = pci_dev->device;

    if (irq)
        *irq = pci_dev->irq;

    return 1;
}

int __ke_pci_get_irq(struct pci_dev *dev, unsigned int* irq)
{
	if (!dev)
		return 0;
	if (!irq)
		return 0;
	*irq = dev->irq;
	return 1;
}

struct pci_dev *__ke_pci_find_device (unsigned int vendor, unsigned int dev, struct pci_dev *from)
{
	return pci_find_device( vendor, dev, from );
}

void* __ke_malloc(__ke_size_t size)
{
    return kmalloc(size, GFP_KERNEL);
}

void __ke_free_s(void* p, __ke_size_t size)
{
    kfree(p);
}

void* __ke_vmalloc(__ke_size_t size)
{
    return vmalloc(size);
}

void __ke_vfree(void* p)
{
    return vfree(p);
}

void* __ke_get_free_page(void)
{
    return (void*)__get_free_page(GFP_KERNEL);
}

void* __ke_get_free_pages(int order)
{
    return (void*)__get_free_pages(GFP_KERNEL, order);
}

void __ke_free_page(void* pt)
{
    free_page((unsigned long)pt);
}

void __ke_free_pages(void* pt, int order)
{
    free_pages((unsigned long)pt, order);
}

void __ke_mem_map_reserve(void* pt)
{
#if LINUX_VERSION_CODE < 0x020400
    mem_map_reserve(MAP_NR((unsigned long)pt));
#else
    mem_map_reserve(virt_to_page((unsigned long)pt));
#endif
}

void __ke_mem_map_unreserve(void* pt)
{
#if LINUX_VERSION_CODE < 0x020400
    mem_map_unreserve(MAP_NR((unsigned long)pt));
#else
    mem_map_unreserve(virt_to_page((unsigned long)pt));
#endif
}

void __ke_virt_reserve(void* virt)
{
#if LINUX_VERSION_CODE < 0x020400
    set_bit(PG_reserved,
        &mem_map[MAP_NR(virt)].flags);
#else
    set_bit(PG_reserved,
        &virt_to_page(virt)->flags);
#endif
}

void __ke_virt_unreserve(void* virt)
{
#if LINUX_VERSION_CODE < 0x020400
    clear_bit(PG_reserved,
        &mem_map[MAP_NR(virt)].flags);
#else
    clear_bit(PG_reserved,
        &virt_to_page(virt)->flags);
#endif
}
                                
void* __ke_ioremap(unsigned long offset, unsigned long size)
{
    return ioremap(offset, size);
}

void* __ke_ioremap_nocache(unsigned long offset, unsigned long size)
{
    return ioremap_nocache(offset, size);
}

void __ke_iounmap(void* pt)
{
    iounmap(pt);
}

int __ke_verify_read_access(void* addr, __ke_size_t size)
{
    return access_ok(VERIFY_READ, addr, size) ? 0 : -EFAULT;
}

int __ke_verify_write_access(void* addr, __ke_size_t size)
{
    return access_ok(VERIFY_WRITE, addr, size) ? 0 : -EFAULT;
}

struct mm_struct* __ke_init_mm(void)
{
    return &init_mm;
}

struct mm_struct* __ke_current_mm(void)
{
    return current->mm;
}

unsigned long __ke_get_phys_addr(struct mm_struct* mm, unsigned long virtual_addr)
{
    unsigned long pte_linear;
    pgd_t* pgd_p;
    pmd_t* pmd_p;
    pte_t* pte_p;
    pte_t  pte;

    pte_linear = VMALLOC_VMADDR(virtual_addr);  // convert to pte linear address (x86 => nop)
    pgd_p = pgd_offset(mm, pte_linear);
    pmd_p = pmd_offset(pgd_p, pte_linear);
#ifndef FGL_ATOMIC_PTE
    pte_p = pte_offset(pmd_p, pte_linear);
    pte = *pte_p;       // duplicate object contents
#else
    pte_p = pte_offset_atomic(pmd_p, pte_linear);
    pte = *pte_p;       // duplicate object contents
    pte_kunmap(pte_p);  // release atomic lock on pte object
#endif

#if LINUX_VERSION_CODE >= 0x020400
    return pte.pte_low & PAGE_MASK;
#else /* LINUX_VERSION_CODE < 0x020400 */
    return virt_to_phys((void*)pte_page(pte_p));
#endif /* LINUX_VERSION_CODE < 0x020400 */
}

void* __ke_memset(void* s, int c, __ke_size_t count)
{
    return memset(s, c, count);
}

void* __ke_memcpy(void* d, const void* s, __ke_size_t count)
{
    return memcpy(d, s, count);
}

__ke_size_t __ke_strlen(const char *s)
{
    return strlen(s);
}

char* __ke_strcpy(char* d, const char* s)
{
    return strcpy(d, s);
}

char* __ke_strncpy(char* d, const char* s, __ke_size_t count)
{
    return strncpy(d, s, count);
}

int __ke_strcmp(const char* string1, const char* string2)
{
    return strcmp(string1, string2);
}

int __ke_strncmp(const char* string1, const char* string2, __ke_size_t count)
{
    return strncmp(string1, string2, count);
}

int __ke_sprintf(char* buf, const char* fmt, ...)
{
    va_list marker;

    va_start(marker, fmt);
    vsprintf(buf, fmt, marker);
    va_end(marker);

    return strlen(buf);
}

/*****************************************************************************/

void __ke_set_bit(int nr, volatile void * addr)
{
    set_bit(nr, addr);
}

void __ke_clear_bit(int nr, volatile void * addr)
{
    clear_bit(nr, addr);
}

/*****************************************************************************/

#ifdef __SMP__
static atomic_t cpus_waiting;

static void deferred_flush(void* contextp)
{
	__asm volatile ("wbinvd":::"memory");
	atomic_dec(&cpus_waiting);
	while (atomic_read(&cpus_waiting) > 0)
		barrier();
}
#endif /* __SMP__ */

int __ke_flush_cache(void)
{
#ifdef __SMP__
	atomic_set(&cpus_waiting, smp_num_cpus - 1);

    /* write back invalidate all other CPUs (exported by kernel) */
	if (smp_call_function(deferred_flush, NULL, 1, 0) != 0)
		panic("timed out waiting for the other CPUs!\n");

    /* invalidate this CPU */
	__asm volatile ("wbinvd":::"memory");

	while (atomic_read(&cpus_waiting) > 0)
		barrier();
#else /* !__SMP__ */
    __asm__ __volatile__ ("wbinvd");
#endif /* !__SMP__ */

    return 0;
}

/*****************************************************************************/

int __ke_config_mtrr(void)
{
#ifdef CONFIG_MTRR
    return 1;
#else /* !CONFIG_MTRR */
    return 0;
#endif /* !CONFIG_MTRR */
}

int __ke_mtrr_add_wc(unsigned long base, unsigned long size)
{
#ifdef CONFIG_MTRR
    return mtrr_add(base, size, MTRR_TYPE_WRCOMB, 1);
#else /* !CONFIG_MTRR */
    return -EPERM;
#endif /* !CONFIG_MTRR */
}

int __ke_mtrr_del(int reg, unsigned long base, unsigned long size)
{
#ifdef CONFIG_MTRR
    return mtrr_del(reg, base, size);
#else /* !CONFIG_MTRR */
    return -EPERM;
#endif /* !CONFIG_MTRR */
}

/*****************************************************************************/

int __ke_pci_read_config_byte(struct pci_dev *dev, __ke_u8 where, __ke_u8 *val)
{
    return pci_read_config_byte(dev, where, val);
}

int __ke_pci_read_config_word(struct pci_dev *dev, __ke_u8 where, __ke_u16 *val)
{
    return pci_read_config_word(dev, where, val);
}

int __ke_pci_read_config_dword(struct pci_dev *dev, __ke_u8 where, __ke_u32 *val)
{
    return pci_read_config_dword(dev, where, val);
}

int __ke_pci_write_config_byte(struct pci_dev *dev, __ke_u8 where, __ke_u8 val)
{
    return pci_write_config_byte(dev, where, val);
}

int __ke_pci_write_config_word(struct pci_dev *dev, __ke_u8 where, __ke_u16 val)
{
    return pci_write_config_word(dev, where, val);
}

int __ke_pci_write_config_dword(struct pci_dev *dev, __ke_u8 where, __ke_u32 val)
{
    return pci_write_config_dword(dev, where, val);
}

/*****************************************************************************/

void __ke_outb(unsigned char value, unsigned short port)
{
    outb(value, port);
}

void __ke_outw(unsigned short value, unsigned short port)
{
    outw(value, port);
}

void __ke_outl(unsigned int value, unsigned short port)
{
    outl(value, port);
}

char __ke_inb(unsigned short port)
{
    return inb(port);
}

short __ke_inw(unsigned short port)
{
    return inw(port);
}

int __ke_inl(unsigned short port)
{
    return inl(port);
}

/*****************************************************************************/
// Interrupt support

void __ke_dbg_irq( void )
{
//	__ke_printk( "__ke_dbg_irq\n");

#if 1
	outl( 0xdeadbeef, 0x80 );
	cli();
	do
	{
	}
	while (1);
#endif
}

void __ke_enable_irq( int irq )
{
	enable_irq( irq );
}

void __ke_disable_irq( int irq )
{
	disable_irq( irq );
}


int __ke_request_irq( unsigned int irq, void (*handler)(int, void *, void *), const char *dev_name, void *dev_id)
{
	return (request_irq( irq, (void(*)(int, void *, struct pt_regs *))handler,
														SA_SHIRQ, dev_name, dev_id ));
}

void __ke_free_irq( unsigned int irq, void *dev_id )
{
	free_irq(irq, dev_id);
}

static struct tq_struct tq;

int __ke_queue_task( void *task, void *data )
{

	tq.routine = task;
	tq.data = data;

	queue_task( &tq, &tq_immediate );
    return 1;
}

void __ke_mark_bh( void )
{
	mark_bh(IMMEDIATE_BH);
}


/*****************************************************************************/

#ifndef NOPAGE_SIGBUS
#define NOPAGE_SIGBUS 0
#endif /* !NOPAGE_SIGBUS */

#if LINUX_VERSION_CODE >= 0x020400
typedef mem_map_t* vm_nopage_ret_t;
#else /* LINUX_VERSION_CODE < 0x020400 */
typedef unsigned long vm_nopage_ret_t;

#endif /* LINUX_VERSION_CODE < 0x020400 */

static vm_nopage_ret_t vm_nopage(struct vm_area_struct* vma,
                                 unsigned long address,
                                 int write_access)
{
    __KE_DEBUG3("start=0x%08lx, "
                "end=0x%08lx, "
                "offset=0x%08lx "
                "address=0x%08lx "
                "write_access=%d\n",
                vma->vm_start,
                vma->vm_end,
                (unsigned long)__ke_vm_offset(vma),
                address,
                write_access);
    return 0;   /* Disallow mremap */
}

/*

    This function is called when a page of a mmap()'ed area is not currently
    visible in the specified VMA.
    Return value is the associated physical address for the requested page.
    (If not implemented, then the kernel default routine would allocate a new,
     zeroed page for servicing us)

    Possible errors: SIGBUS, OutOfMem

    This routine is intended to remap addresses of SHM SAREA
    (which is one or more pages in size)

 */
static vm_nopage_ret_t vm_shm_nopage(struct vm_area_struct* vma,
                                     unsigned long address,
                                     int write_access)
{
    pgd_t* pgd_p;
    pmd_t* pmd_p;
    pte_t* pte_p;
    pte_t  pte;
    unsigned long vma_offset;
    unsigned long linear;
    unsigned long pte_linear;
#if LINUX_VERSION_CODE < 0x020400
    unsigned long kaddr;
#endif /*  LINUX_VERSION_CODE < 0x020400 */
    mem_map_t* pMmPage;

    __KE_DEBUG3("start=0x%08lx, "
                "end=0x%08lx, "
                "offset=0x%08lx "
                "address=0x%08lx "
                "write_access=%d\n",
                vma->vm_start,
                vma->vm_end,
                __ke_vm_offset(vma),
                address,
                write_access);

    /*
        vm_start           => start of vm-area,  regular address
        vm_end             => end of vm-area,    regular address
        vm_offset/vm_pgoff => start of area,     linear address
        address            => requested address, regular address

        Check range
        Seems the surrounding framework already does that test -
        skip it here, anyone does.
     */
#if 0
    if (address < vma->vm_start)
        return NOPAGE_SIGBUS; /* Address is out of range */
#endif
    /*
        Note: vm_end is not member of range but this border
        hmm, might be used when growing the VMA, not sure - keep it as it is.
     */
    if (address > vma->vm_end)
        return NOPAGE_SIGBUS; /* address is out of range */

    /*  Calculate offset into VMA */
    vma_offset = address - vma->vm_start;

    /*
        TRICK:
        We dont have to keep a correspondence list between VMA offsets and
        physiucal address. But since that memory is mapped into kernel space
        we just have to climb down the kernel page table structures to get
        the matching kernel (mem_map/page) address for the given offset.
        This works because memory is created in kernel memory and therefore
        it is just present even if its not yet mapped to current VMA.

        Sample: 32 MB adapter RAM -> 8192 pages a 4kB -> 32 kB Data on x86

        Translate VMA offset into a linear address (=virtual address),
        at kernel scope, by simply adding the base (=VM_OFFSET) address of
        the VMA in kernel space.
     */
    linear = __ke_vm_offset(vma) + vma_offset;

    /*
        For historical reasons:
        pte used some offset on x86 kernels previous to 2.1.1

        Convert to pte linear address (x86 => nop)
     */
    pte_linear = VMALLOC_VMADDR(linear);

    /*
        Locate responsible kernel PTE for this linear address
        in paging system of the kernel VM
        (or is it the init process? - not sure yet)
     */
    pgd_p = pgd_offset(&init_mm, pte_linear); /* locate page directory entry */
    if (!pgd_present(*pgd_p))
    {
        __KE_ERROR("FATAL ERROR: User queue buffer not present! (pgd)\n");
        return NOPAGE_SIGBUS;   /* Something bad happened; generate SIGBUS */
        /* alternatively we could generate a NOPAGE_OOM "out of memory" */
    }
    /*  locate medium level page table (x86 => nop) */
    pmd_p = pmd_offset(pgd_p, pte_linear);
    if (!pmd_present(*pmd_p))
    {
        __KE_ERROR("FATAL ERROR: User queue buffer not present! (pmd)\n");
        return NOPAGE_SIGBUS;   /* Something bad happened; generate SIGBUS */
        /* alternatively we could generate a NOPAGE_OOM "out of memory" */
    }
    /*  locate page table entry itself */
#ifndef FGL_ATOMIC_PTE
    pte_p = pte_offset(pmd_p, pte_linear);
    pte = *pte_p;       // duplicate contents
#else
    pte_p = pte_offset_atomic(pmd_p, pte_linear);
    pte = *pte_p;       // duplicate contents
    pte_kunmap(pte_p);  // release atomic lock
#endif
    if (!pte_present(pte))
    {
        __KE_ERROR("FATAL ERROR: User queue buffer not present! (pte)\n");
        return NOPAGE_SIGBUS;   /* Something bad happened; generate SIGBUS */
        /* alternatively we could generate a NOPAGE_OOM "out of memory" */
    }

    /*
        Resolve the kernel (mem_map/page) address for the VMA-address
        we got queried about.
    */
#if LINUX_VERSION_CODE >= 0x020400
    /*
        Pretty straight in new kernel. The pte is represented by the requested
        pointer to the page table entry.
     */
    pMmPage = pte_page(pte);
#else /* LINUX_VERSION_CODE < 0x020400 */
    kaddr = pte_page(pte);
    pMmPage = virt_to_page(kaddr);
#endif /* LINUX_VERSION_CODE < 0x020400 */

    atomic_inc(&(pMmPage->count));  /* inc usage count of page */

#if LINUX_VERSION_CODE >= 0x020400
    __KE_DEBUG3("vm-address 0x%08lx => kernel-page-address 0x%p\n",
        address, page_address(pMmPage));
    return pMmPage;
#else /* LINUX_VERSION_CODE < 0x020400 */
    __KE_DEBUG3("vm-address 0x%08lx => kernel-page-address 0x%p\n",
        address, (void*)kaddr);
    return kaddr;
#endif /* LINUX_VERSION_CODE < 0x020400 */
}

/*

    This routine is intended to remap addresses of a OpenGL context
      (which is one ore more pages in size)

*/
static vm_nopage_ret_t vm_ctx_nopage(struct vm_area_struct* vma,
                                     unsigned long address,
                                     int write_access)
{
    unsigned long kaddr;
    mem_map_t* pMmPage;

    __KE_DEBUG3("start=0x%08lx, "
                "end=0x%08lx, "
                "offset=0x%08lx "
                "address=0x%08lx "
                "write_access=%d\n",
                vma->vm_start,
                vma->vm_end,
                __ke_vm_offset(vma),
                address,
                write_access);

    if (address > vma->vm_end)
        return 0; /* Disallow mremap */

    /*
        Have we ever got an acces from user land into context structure?

        Resolve the kernel (mem_map/page) address for the VMA-address
        we got queried about.
    */
    kaddr = __ke_vm_offset(vma) + (address - vma->vm_start);
    pMmPage = virt_to_page(kaddr);

#if 0
    // WARNING WARNINIG WARNNING WARNNING WARNNING WARNNING WARNNING WARNNING
    // Don't increment page usage count, cause ctx pages are allocated
    // with drm_alloc_pages, which marks all pages as reserved. Reserved
    // pages' usage count is not decremented by the kernel during unmap!!!
    atomic_inc(&(pMmPage->count)); /* inc usage count of page */
#endif

#if LINUX_VERSION_CODE >= 0x020400
    __KE_DEBUG3("vm-address 0x%08lx => kernel-page-address 0x%p\n",
        address, page_address(pMmPage));
    return pMmPage;
#else /*  LINUX_VERSION_CODE < 0x020400 */
    __KE_DEBUG3("vm-address 0x%08lx => kernel-page-address 0x%p\n",
        address, (void*)kaddr);
    return kaddr;
#endif /* LINUX_VERSION_CODE < 0x020400 */
}

void* __ke_vma_file_priv(struct vm_area_struct* vma)
{
    return vma->vm_file->private_data;
}

unsigned long __ke_vm_start(struct vm_area_struct* vma)
{
    return vma->vm_start;
}

unsigned long __ke_vm_end(struct vm_area_struct* vma)
{
    return vma->vm_end;
}

unsigned long __ke_vm_offset(struct vm_area_struct* vma)
{
#if LINUX_VERSION_CODE < 0x020319
    return vma->vm_offset;
#else /* LINUX_VERSION_CODE >= 0x020319 */
    return vma->vm_pgoff << PAGE_SHIFT;
#endif/* LINUX_VERSION_CODE >= 0x020319 */
}

char* __ke_vm_flags_str(struct vm_area_struct* vma, char* buf)
{
   *(buf + 0) = vma->vm_flags & VM_READ	    ? 'r' : '-';
   *(buf + 1) = vma->vm_flags & VM_WRITE	? 'w' : '-';
   *(buf + 2) = vma->vm_flags & VM_EXEC	    ? 'x' : '-';
   *(buf + 3) = vma->vm_flags & VM_MAYSHARE ? 's' : 'p';
   *(buf + 4) = vma->vm_flags & VM_LOCKED   ? 'l' : '-';
   *(buf + 5) = vma->vm_flags & VM_IO	    ? 'i' : '-';
   *(buf + 6) = 0;
   return buf;
}

char* __ke_vm_page_prot_str(struct vm_area_struct* vma, char* buf)
{
    int i = 0;

#ifdef __i386__
	unsigned int pgprot;

    pgprot = pgprot_val(vma->vm_page_prot);
    *(buf + i++) = pgprot & _PAGE_PRESENT  ? 'p' : '-';
    *(buf + i++) = pgprot & _PAGE_RW       ? 'w' : 'r';
    *(buf + i++) = pgprot & _PAGE_USER     ? 'u' : 's';
    *(buf + i++) = pgprot & _PAGE_PWT      ? 't' : 'b';
    *(buf + i++) = pgprot & _PAGE_PCD      ? 'u' : 'c';
    *(buf + i++) = pgprot & _PAGE_ACCESSED ? 'a' : '-';
    *(buf + i++) = pgprot & _PAGE_DIRTY    ? 'd' : '-';
#if LINUX_VERSION_CODE >= 0x020400
    *(buf + i++) = pgprot & _PAGE_PSE      ? 'm' : 'k';
#else /* LINUX_VERSION_CODE < 0x020400 */
    *(buf + i++) = pgprot & _PAGE_4M       ? 'm' : 'k';
#endif /* LINUX_VERSION_CODE < 0x020400 */
    *(buf + i++) = pgprot & _PAGE_GLOBAL   ? 'g' : 'l';
#endif /* __i386__ */		
    *(buf + i++) = 0;

    return buf;
}

char* __ke_vm_phys_addr_str(struct vm_area_struct* vma, 
                            char* buf, 
                            unsigned long virtual_addr, 
                            unsigned long* phys_address)
{
    unsigned long pte_linear;
    pgd_t* pgd_p;
    pmd_t* pmd_p;
    pte_t* pte_p;
    pte_t  pte;

    pte_linear = VMALLOC_VMADDR(virtual_addr);  // convert to pte linear address (x86 => nop)
    pgd_p = pgd_offset(vma->vm_mm, pte_linear);
    pmd_p = pmd_offset(pgd_p, pte_linear);
#ifndef FGL_ATOMIC_PTE
    pte_p = pte_offset(pmd_p, pte_linear);
    pte = *pte_p;
#else
    pte_p = pte_offset_atomic(pmd_p, pte_linear);
    pte = *pte_p;
    pte_kunmap(pte_p);  // release atomic lock
#endif

	if (pte_present(pte))
    {
#if LINUX_VERSION_CODE >= 0x020400
        *phys_address = pte.pte_low & PAGE_MASK;
#else /* LINUX_VERSION_CODE < 0x020400 */
        *phys_address = virt_to_phys((void*)pte_page(pte));
#endif /* LINUX_VERSION_CODE < 0x020400 */

        sprintf(buf, "0x%08lx %c%c%c%c%c\n",
           *phys_address,                                                        
           pte_read (pte) ? 'r' : '-',                                           
           pte_write(pte) ? 'w' : '-',                                           
           pte_exec (pte) ? 'x' : '-',                                           
           pte_dirty(pte) ? 'd' : '-',                                           
           pte_young(pte) ? 'a' : '-');
    }
    else
        *buf = 0;

    return buf;
}

static struct vm_operations_struct vm_ops =
{
    nopage:  vm_nopage,
    open:    drm_vm_open,
    close:   drm_vm_close,
};

static struct vm_operations_struct vm_shm_ops =
{
    nopage:  vm_shm_nopage,
    open:    drm_vm_open,
    close:   drm_vm_close,
};

static struct vm_operations_struct vm_pci_bq_ops =
{
    nopage:  vm_shm_nopage,
    open:    firegl_vm_bq_open,
    close:   firegl_vm_bq_close,
};

static struct vm_operations_struct vm_ctx_ops =
{
    nopage:  vm_ctx_nopage,
    open:    drm_vm_ctx_open,
    close:   drm_vm_ctx_close,
};

// always defined
#define __AGP__BUILTIN__

#ifdef __AGP__BUILTIN__
static struct vm_operations_struct vm_agp_bq_ops =
{
    nopage:  vm_nopage,
    open:    firegl_vm_bq_open,
    close:   firegl_vm_bq_close,
};
#endif /* __AGP__BUILTIN__ */

int __ke_vm_map(struct file* filp,
                struct vm_area_struct* vma,
                enum __ke_vm_maptype type,
                int readonly)
{
    unsigned int pages;

	__KE_VMDEBUG(" type <0x%x>, handle <0x%x>\n", type, filp);


    switch (type)
    {
        case __KE_ADPT:
            if (__ke_vm_offset(vma) >= __pa(high_memory))
            {
#ifdef __i386__
                if (boot_cpu_data.x86 > 3)
                {
                    pgprot_val(vma->vm_page_prot) |= _PAGE_PCD;
                    pgprot_val(vma->vm_page_prot) &= ~_PAGE_PWT;
                }
#endif /* __i386__ */
                vma->vm_flags |= VM_IO; /* not in core dump */
            }
            if (remap_page_range(FGL_VMA_API_PASS
                                 vma->vm_start,
                                 __ke_vm_offset(vma),
                                 vma->vm_end - vma->vm_start,
                                 vma->vm_page_prot))
            {
                __KE_DEBUG("remap_page_range failed\n");
                return -EAGAIN;
            }
            vma->vm_flags |= VM_SHM; /* Don't swap */
            vma->vm_ops = &vm_ops;
            break;

        case __KE_SHM:
            vma->vm_flags |= VM_SHM; /* Don't swap */
            vma->vm_ops = &vm_shm_ops;
            break;

        case __KE_CTX:
            pages = (vma->vm_end - vma->vm_start) >> PAGE_SHIFT;
            vma->vm_flags |= VM_LOCKED | VM_SHM; /* Don't swap */
            vma->vm_mm->locked_vm += pages; /* Kernel tracks aqmount of locked pages */
            vma->vm_ops = &vm_ctx_ops;
            break;

        case __KE_PCI_BQS:
            pages = (vma->vm_end - vma->vm_start) >> PAGE_SHIFT;
            vma->vm_flags |= VM_LOCKED | VM_SHM; /* Don't swap */
            vma->vm_mm->locked_vm += pages; /* Kernel tracks aqmount of locked pages */
            vma->vm_ops = &vm_pci_bq_ops;
            break;

#ifdef __AGP__BUILTIN__
        case __KE_AGP:
            if (__ke_vm_offset(vma) >= __pa(high_memory))
                vma->vm_flags |= VM_IO; /* not in core dump */
            if (remap_page_range(FGL_VMA_API_PASS
                                 vma->vm_start,
                                 __ke_vm_offset(vma),
                                 vma->vm_end - vma->vm_start,
                                 vma->vm_page_prot))
            {
                __KE_DEBUG("remap_page_range failed\n");
                return -EAGAIN;
            }
            vma->vm_flags |= VM_SHM; /* Don't swap */
            vma->vm_ops = &vm_ops;
            break;

        case __KE_AGP_BQS:
            if (__ke_vm_offset(vma) >= __pa(high_memory))
                vma->vm_flags |= VM_IO; /* not in core dump */
            if (remap_page_range(FGL_VMA_API_PASS
                                 vma->vm_start,
                                 __ke_vm_offset(vma),
                                 vma->vm_end - vma->vm_start,
                                 vma->vm_page_prot))
            {
                __KE_DEBUG("remap_page_range failed\n");
                return -EAGAIN;
            }
            vma->vm_flags |= VM_SHM; /* Don't swap */
            vma->vm_ops = &vm_agp_bq_ops;
            break;
#endif /* __AGP__BUILTIN__ */

        default:
            /*  This should never happen anyway! */
            __KE_ERROR("__ke_vm_map: Unknown type %d\n", type);
            return -EINVAL;
    }

    if (readonly)
        pgprot_val(vma->vm_page_prot) &= ~_PAGE_RW;

#if LINUX_VERSION_CODE < 0x020203 /* KERNEL_VERSION(2,2,3) */
    /* In Linux 2.2.3 and above, this is handled in do_mmap() in mm/mmap.c. */
    ++filp->f_count;
#endif /* LINUX_VERSION_CODE < 0x020203 */
    vma->vm_file = filp;    /* Needed for drm_vm_open() */

    return 0;
}


/*****************************************************************************/

#ifdef __AGP__

#if LINUX_VERSION_CODE >= 0x020400

static const drm_agp_t* drm_agp = NULL;

#define AGP_FUNCTIONS		8
#define AGP_AVAILABLE(func)	(drm_agp && drm_agp-> func )
#define AGP_FUNC(func)		(*drm_agp-> func )
// note: avoid ##-tokens with latest GCC (i.e. RedHat 7.1 beta)

#else /* LINUX_VERSION_CODE < 0x020400 */
enum agp_func_ndx {
    __ndx_agp_free_memory,
    __ndx_agp_allocate_memory,
    __ndx_agp_bind_memory,
    __ndx_agp_unbind_memory,
    __ndx_agp_enable,
    __ndx_agp_acquire,
    __ndx_agp_release,
    __ndx_agp_copy_info,
};

/* The C standard says that 'void *' is not guaranteed to hold a function
   pointer, so we use this union to define a generic pointer that is
   guaranteed to hold any of the function pointers we care about. */
typedef union 
{
	void (*free_memory)(agp_memory*);
	agp_memory *(*allocate_memory)(size_t, u32);
	int (*bind_memory)(agp_memory*, off_t);
	int (*unbind_memory)(agp_memory*);
	void (*enable)(u32);
	int (*acquire)(void);
	void (*release)(void);
	void (*copy_info)(agp_kern_info*);
	unsigned long address;
} agp_func_u;

typedef struct agp_fill
{
    const char* name;   /*  pointer to the name of the symbol to resolve */
    agp_func_u f;       /*  pointer to function from agp_gart-module */
} agp_fill_t;

static agp_fill_t drm_agp[] = {
    [__ndx_agp_free_memory] = 
        { __MODULE_STRING(agp_free_memory),		{ address: 0 } },
    [__ndx_agp_allocate_memory] = 
        { __MODULE_STRING(agp_allocate_memory), { address: 0 } },
    [__ndx_agp_bind_memory] = 
        { __MODULE_STRING(agp_bind_memory),     { address: 0 } },
    [__ndx_agp_unbind_memory] = 
        { __MODULE_STRING(agp_unbind_memory),   { address: 0 } },
    [__ndx_agp_enable] = 
        { __MODULE_STRING(agp_enable),          { address: 0 } },
    [__ndx_agp_acquire] = 
        { __MODULE_STRING(agp_backend_acquire), { address: 0 } },
    [__ndx_agp_release] = 
        { __MODULE_STRING(agp_backend_release), { address: 0 } },
    [__ndx_agp_copy_info] = 
        { __MODULE_STRING(agp_copy_info),       { address: 0 } },
};

#define AGP_FUNCTIONS		(sizeof(drm_agp) / sizeof(drm_agp[0]))
#define AGP_AVAILABLE(func)	(drm_agp[__ndx_agp_##func].f.address != 0)
#define AGP_FUNC(func)		(*drm_agp[__ndx_agp_##func].f. func)

#endif /* LINUX_VERSION_CODE < 0x020400 */

#endif /* __AGP__ */

static int firegl_agp = 0;
int __ke_agp_try_unsupported = 0;

// WARNING: Don't enable USE_FIREGL_AGPGART_IMPLEMENTATION, because it's not yet implemented!!!
//#define USE_FIREGL_AGPGART_IMPLEMENTATION
#ifdef USE_FIREGL_AGPGART_IMPLEMENTATION
#define FIREGL_agp_init             firegl_agp_init
#define FIREGL_agp_cleanup          firegl_agp_cleanup
#define FIREGL_agp_free_memory      firegl_agp_free_memory
#define FIREGL_agp_allocate_memory  firegl_agp_allocate_memory
#define FIREGL_agp_bind_memory      firegl_agp_bind_memory
#define FIREGL_agp_unbind_memory    firegl_agp_unbind_memory
#define FIREGL_agp_enable           firegl_agp_enable
#define FIREGL_agp_backend_acquire  firegl_agp_acquire
#define FIREGL_agp_backend_release  firegl_agp_release
#define FIREGL_agp_memory           struct _agp_memory
#else // !USE_FIREGL_AGPGART_IMPLEMENTATION
#define FIREGL_agp_init             _X(agp_init)
#define FIREGL_agp_cleanup          _X(agp_cleanup)
#define FIREGL_agp_free_memory      _X(agp_free_memory)
#define FIREGL_agp_allocate_memory  _X(agp_allocate_memory)
#define FIREGL_agp_bind_memory      _X(agp_bind_memory)
#define FIREGL_agp_unbind_memory    _X(agp_unbind_memory)
#define FIREGL_agp_enable           _X(agp_enable)
#define FIREGL_agp_backend_acquire  _X(agp_backend_acquire)
#define FIREGL_agp_backend_release  _X(agp_backend_release)
#define FIREGL_agp_memory           _X(agp_memory)
#endif // !USE_FIREGL_AGPGART_IMPLEMENTATION

static
int __ke_firegl_agpgart_available(void)
{
    int retval;

#ifndef USE_FIREGL_AGPGART_IMPLEMENTATION
    if (__ke_agp_try_unsupported)
        _X(agp_try_unsupported) = __ke_agp_try_unsupported;
#endif // !USE_FIREGL_AGPGART_IMPLEMENTATION

    retval = FIREGL_agp_init();
    if (retval) 
    {
        __KE_DEBUG("Initialization of built-in AGP-support failed (ret=%d).\n", retval);
        return 0;
    }

    __KE_DEBUG("Initialization of built-in AGP-support successful.\n");
    firegl_agp = 1;

    return 1;
}

static
int __ke_agpgart_available(void)
{
#ifdef __AGP__
    unsigned int found = 0;
    
    __KE_DEBUG("\n");

#if LINUX_VERSION_CODE >= 0x020400
	drm_agp = (drm_agp_t*)inter_module_get("drm_agp");
    if (drm_agp) 
    {
	    if (drm_agp->free_memory) 
	    {
            __KE_DEBUG("agp_free_memory resolves to 0x%08lx\n", drm_agp->free_memory);
		    found++;
	    }
	    if (drm_agp->allocate_memory) 
	    {
            __KE_DEBUG("agp_allocate_memory resolves to 0x%08lx\n", drm_agp->allocate_memory);
		    found++;
	    }
	    if (drm_agp->bind_memory) 
	    {
            __KE_DEBUG("agp_bind_memory resolves to 0x%08lx\n", drm_agp->bind_memory);
		    found++;
	    }
	    if (drm_agp->unbind_memory) 
	    {
            __KE_DEBUG("agp_unbind_memory resolves to 0x%08lx\n", drm_agp->unbind_memory);
		    found++;
	    }
	    if (drm_agp->enable) 
	    {
            __KE_DEBUG("agp_enable resolves to 0x%08lx\n", drm_agp->enable);
		    found++;
	    }
	    if (drm_agp->acquire) 
	    {
            __KE_DEBUG("agp_acquire resolves to 0x%08lx\n", drm_agp->acquire);
		    found++;
	    }
	    if (drm_agp->release) 
	    {
            __KE_DEBUG("agp_release resolves to 0x%08lx\n", drm_agp->release);
		    found++;
	    }
	    if (drm_agp->copy_info) 
	    {
            __KE_DEBUG("agp_copy_info resolves to 0x%08lx\n", drm_agp->copy_info);
		    found++;
	    }
    }
#else /* LINUX_VERSION_CODE < 0x020400 */
	{
	    int i;

		for (i = 0; i < sizeof(drm_agp) / sizeof(drm_agp[0]); i++)
		{
			drm_agp[i].f = 
				(agp_func_u)get_module_symbol(NULL, (char*)drm_agp[i].name);
			__KE_DEBUG("%s resolves to 0x%08lx\n", 
				drm_agp[i].name, drm_agp[i].f.address);
			if (drm_agp[i].f.address)
				found++;
		}
	}
#endif /* LINUX_VERSION_CODE < 0x020400 */
    
    if (found == AGP_FUNCTIONS) {
        __KE_DEBUG("AGP/GART kernel module present, API is complete.\n");
        return 1;
    }

    /* do a complete cleanup of our entry point table now */
    __ke_agp_uninit();

    if (found == 0)
        __KE_DEBUG("AGP/GART kernel module not found\n");
    else
        __KE_DEBUG("AGP/GART kernel module present but API is incomplete\n");

#else /* !__AGP__ */
    __KE_DEBUG("Kernel native AGP support was not present at module build time.\n");
#endif /* !__AGP__ */

    return 0;
}

int __ke_agp_available(void)
{
    int available = 0;

    if (!(__ke_moduleflags & __KE_FLAG_DISABLE_FIREGL_AGPGART))
        available = __ke_firegl_agpgart_available();

    if (!available && !(__ke_moduleflags & __KE_FLAG_DISABLE_AGPGART))
        available = __ke_agpgart_available();

    return available;
}

void __ke_agp_uninit(void)
{
    if (firegl_agp) 
    {
        FIREGL_agp_cleanup();
        firegl_agp = 0;
        return;
    }

#ifdef __AGP__
#if LINUX_VERSION_CODE >= 0x020400
    if (drm_agp)
	    inter_module_put("drm_agp");
	drm_agp = NULL;
#else /* LINUX_VERSION_CODE < 0x020400 */
    {
        int i;
        for (i = 0; i < sizeof(drm_agp) / sizeof(drm_agp[0]); i++)
        {
            drm_agp[i].f.address = 0;
        }
    }
#endif /* LINUX_VERSION_CODE < 0x020400 */
#endif /* __AGP__ */
}

void __ke_agp_free_memory(struct _agp_memory* handle)
{
#ifdef __AGP__
	if (AGP_AVAILABLE(free_memory))
		AGP_FUNC(free_memory)(handle);
#endif /* __AGP__ */
	if (firegl_agp)
		return FIREGL_agp_free_memory((FIREGL_agp_memory*)handle);
}

struct _agp_memory* __ke_agp_allocate_memory(size_t pages, unsigned long type)
{
#ifdef __AGP__
	if (AGP_AVAILABLE(allocate_memory))
		return AGP_FUNC(allocate_memory)(pages, type);
#endif /* __AGP__ */
	if (firegl_agp)
		return (struct _agp_memory*)FIREGL_agp_allocate_memory(pages, type);
    return NULL;
}

int __ke_agp_bind_memory(struct _agp_memory* handle, __ke_off_t start)
{
#ifdef __AGP__
	if (AGP_AVAILABLE(bind_memory))
		return AGP_FUNC(bind_memory)(handle, start);
#endif /* __AGP__ */
	if (firegl_agp)
		return FIREGL_agp_bind_memory((FIREGL_agp_memory*)handle, start);
    return -EINVAL;
}

int __ke_agp_unbind_memory(struct _agp_memory* handle)
{
#ifdef __AGP__
	if (AGP_AVAILABLE(unbind_memory))
		return AGP_FUNC(unbind_memory)(handle);
#endif /* __AGP__ */
	if (firegl_agp)
		return FIREGL_agp_unbind_memory((FIREGL_agp_memory*)handle);
    return -EINVAL;
}

int __ke_agp_enable(unsigned long mode)
{
#ifdef __AGP__
	if (AGP_AVAILABLE(enable))
    {
		AGP_FUNC(enable)(mode);
        return 0;
    }
#endif /* __AGP__ */
	if (firegl_agp)
    {
		FIREGL_agp_enable(mode);
        return 0;
    }
    return -EINVAL;
}

int __ke_read_agp_command_register(struct pci_dev *pdevice, unsigned long *cmd)
{
    u8 capndx;
	u32 cap_id;
    u32 command;
    if ( pdevice && cmd )
    {
	    pci_read_config_byte(pdevice, 0x34, &capndx);
	    if (capndx != 0x00)
        {
            do { // search capability list
			    pci_read_config_dword(pdevice, capndx, &cap_id);
			    if ((cap_id & 0xff) != 0x02)
				    capndx = (cap_id >> 8) & 0xff;
            } while (((cap_id & 0xff) != 0x02) && (capndx != 0x00));
            pci_read_config_dword(pdevice, capndx + 8, &command);
            *cmd = command;
            return 0;
	    }
    }
    return -EINVAL;
}

int __ke_agp_acquire(void)
{
#ifdef __AGP__
	if (AGP_AVAILABLE(acquire))
		return AGP_FUNC(acquire)();
#endif /* __AGP__ */
	if (firegl_agp)
		return FIREGL_agp_backend_acquire();
    return -EINVAL;
}

void __ke_agp_release(void)
{
#ifdef __AGP__
	if (AGP_AVAILABLE(release))
		AGP_FUNC(release)();
#endif /* __AGP__ */
	if (firegl_agp)
		FIREGL_agp_backend_release();
}

void __ke_agp_copy_info(__ke_agp_kern_info_t* info)
{
    memset(info, 0, sizeof(__ke_agp_kern_info_t));

#ifdef __AGP__
	if (AGP_AVAILABLE(copy_info))
    {
        agp_kern_info kern;    

		AGP_FUNC(copy_info)(&kern);

        info->version.major = kern.version.major;
        info->version.minor = kern.version.minor;
        info->vendor = kern.device->vendor;
        info->device = kern.device->device;
        info->mode = kern.mode;
        info->aper_base = kern.aper_base;
        info->aper_size = kern.aper_size;
        info->max_memory = kern.max_memory;
        info->current_memory = kern.current_memory;
    }
#endif /* __AGP__ */
    if (firegl_agp) 
    {
#ifdef USE_FIREGL_AGPGART_IMPLEMENTATION
		firegl_agp_copy_info(info);
#else // !USE_FIREGL_AGPGART_IMPLEMENTATION
        _X(agp_kern_info) kern;    

		_X(agp_copy_info)(&kern);

        info->version.major = kern.version.major;
        info->version.minor = kern.version.minor;
        info->vendor = kern.device->vendor;
        info->device = kern.device->device;
        info->mode = kern.mode;
        info->aper_base = kern.aper_base;
        info->aper_size = kern.aper_size;
        info->max_memory = kern.max_memory;
        info->current_memory = kern.current_memory;
#endif // !USE_FIREGL_AGPGART_IMPLEMENTATION
    }
}

unsigned long __ke_agp_memory_handle(struct _agp_memory* handle)
{
    if (firegl_agp) 
#ifdef USE_FIREGL_AGPGART_IMPLEMENTATION
        return firegl_agp_memory_handle(handle);
#else // !USE_FIREGL_AGPGART_IMPLEMENTATION
        return (unsigned long)((_X(agp_memory)*)handle)->memory;
#endif // !USE_FIREGL_AGPGART_IMPLEMENTATION

#ifdef __AGP__
    return (unsigned long)handle->memory;
#else /* !__AGP__ */
    return 0;
#endif /* !__AGP__ */
}

unsigned long __ke_agp_memory_page_count(struct _agp_memory* handle)
{
    if (firegl_agp) 
#ifdef USE_FIREGL_AGPGART_IMPLEMENTATION
        return firegl_agp_memory_page_count(handle);
#else // !USE_FIREGL_AGPGART_IMPLEMENTATION
        return ((_X(agp_memory)*)handle)->page_count;
#endif // !USE_FIREGL_AGPGART_IMPLEMENTATION

#ifdef __AGP__
    return handle->page_count;
#else /* !__AGP__ */
    return 0;
#endif /* !__AGP__ */
}

int __ke_smp_processor_id(void)
{
	return (int)(smp_processor_id());
}


void __ke_smp_call_function( void (*func)(void *info) )
{
	smp_call_function( func, NULL, 0, 1 );
}

#define vendorStringLen 12

static inline int check_cpu_vendor(const char* vendorName)
{
    int sVendor[vendorStringLen/sizeof(int)];
    __asm
    (
        "\txor %%eax, %%eax\n"  // function #0: get vendor string
        "\tcpuid\n"             // perform CPUID
        "\tmov %%ebx, %0\n"     // store 3 times 4 bytes of vendor string
        "\tmov %%edx, %1\n"
        "\tmov %%ecx, %2\n"
        : "=m" (sVendor[0]), "=m" (sVendor[1]), "=m"(sVendor[2])   // results go to this location
        :                       // no input
        : "eax", "ebx", "ecx", "edx"    // modifiys this registers
    );

    if( strncmp((char*)sVendor,vendorName,vendorStringLen)==0 )
        return 1;   // thats a match
    else
        return 0;   // does not match
}

int __ke_is_athlon(void)
{
    register int bAthlon;

    __asm
    (
        // Check for CPUID instruction
        "\tpushf\n"              	//; save EFLAGS
        "\tpop %%eax\n"             	//; store EFLAGS in EAX
        "\tmov %%eax, %%ebx\n"        	//; save in EBX for later testing
        "\txor $0x00200000, %%eax\n"  	//; toggle bit 21
        "\tpush %%eax\n"            	//; put to stack
        "\tpopf\n"               	//; save changed EAX to EFLAGS
        "\tpushf\n"              	//; push EFLAGS to TOS
        "\tpop %%eax\n"             	//; store EFLAGS in EAX
        "\tcmp %%ebx, %%eax\n"        	//; see if bit 21 has changed
        "\tjz NO_ATHLON\n"        	//; if no change, no CPUID

        // Check for extended functions
        "\tmov $0x80000000, %%eax\n"  	//; query for extended functions
        "\tCPUID\n"               	//; get extended function limit
        "\tcmp $0x80000000, %%eax\n"  	//; is 8000_0001h supported?
        "\tjbe NO_ATHLON\n"       	//; if not, 3DNow! command set not supported

        "\tmov $0x80000001, %%eax\n"  	//; setup extended function 1
        "\tCPUID\n"               	//; call the function
        "\ttest $0x80000000, %%edx\n" 	//; test bit 31
        "\tjz NO_ATHLON\n"        	//; 3DNow! command set not supported
        "\tmov $1, %%eax\n"
        "\tjmp DONE\n"
"NO_ATHLON:\n"
        "\txor %%eax, %%eax\n"
"DONE:\n"
        "\tmov %%eax, %0\n"         //; store to bAthlon
        // optimizer hints
        : "=r" (bAthlon)            // let compiler choose output register
        :                           // no input selection for compiler
        : "eax", "ebx", "ecx", "edx"	// modifiys
    );

    if( bAthlon )
    {
        if( !check_cpu_vendor("AuthenticAMD") )
            bAthlon = 0;
    }

    return bAthlon;
}

#endif /* __KERNEL__ */
